from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field

from grim.geom import Vec2

from ..camera import camera_shake_update
from ..creatures.damage import creature_apply_damage
from ..creatures.runtime import CREATURE_HITBOX_ALIVE, CreatureDeath, CreaturePool
from ..creatures.anim import creature_anim_advance_phase
from ..creatures.spawn import CreatureFlags, CreatureTypeId, SpawnEnv
from ..effects import FxQueue, FxQueueRotated
from ..features.bonuses import emit_bonus_pickup_effects
from ..features.perks import PLAYER_DEATH_HOOKS, WORLD_DT_STEPS
from ..game_modes import GameMode
from ..gameplay import (
    BonusPickupEvent,
    GameplayState,
    PlayerInput,
    PlayerState,
    bonus_update,
    bonus_update_pre_pickup_timers,
    perks_update_effects,
    player_update,
    survival_progression_update,
)
from ..player_damage import player_take_projectile_damage
from ..projectiles import ProjectileHit
from .input_frame import normalize_input_frame
from .presentation_step import (
    ProjectileDecalPostCtx,
    plan_death_sfx_keys,
    plan_hit_sfx_keys,
    queue_projectile_decals_post_hit,
    queue_projectile_decals_pre_hit,
)
from .world_defs import CREATURE_ANIM


@dataclass(slots=True)
class WorldEvents:
    hits: list[ProjectileHit]
    deaths: tuple[CreatureDeath, ...]
    pickups: list[BonusPickupEvent]
    sfx: list[str]
    trigger_game_tune: bool = False
    hit_sfx: list[str] = field(default_factory=list)
    death_sfx_preplanned: bool = False


_WORLD_DT_STEPS = WORLD_DT_STEPS
_PLAYER_DEATH_HOOKS = PLAYER_DEATH_HOOKS


@dataclass(slots=True)
class WorldState:
    spawn_env: SpawnEnv
    state: GameplayState
    players: list[PlayerState]
    creatures: CreaturePool

    @classmethod
    def build(
        cls,
        *,
        world_size: float,
        demo_mode_active: bool,
        hardcore: bool,
        difficulty_level: int,
        preserve_bugs: bool = False,
    ) -> WorldState:
        spawn_env = SpawnEnv(
            terrain_width=float(world_size),
            terrain_height=float(world_size),
            demo_mode_active=bool(demo_mode_active),
            hardcore=bool(hardcore),
            difficulty_level=int(difficulty_level),
        )
        state = GameplayState()
        state.demo_mode_active = bool(demo_mode_active)
        state.hardcore = bool(hardcore)
        state.preserve_bugs = bool(preserve_bugs)
        players: list[PlayerState] = []
        creatures = CreaturePool(env=spawn_env, effects=state.effects)
        return cls(
            spawn_env=spawn_env,
            state=state,
            players=players,
            creatures=creatures,
        )

    def step(
        self,
        dt: float,
        *,
        dt_ms_i32: int | None = None,
        defer_camera_shake_update: bool = False,
        defer_freeze_corpse_fx: bool = False,
        mid_step_hook: Callable[[], None] | None = None,
        inputs: list[PlayerInput] | None,
        world_size: float,
        damage_scale_by_type: dict[int, float],
        detail_preset: int,
        fx_toggle: int = 0,
        fx_queue: FxQueue,
        fx_queue_rotated: FxQueueRotated,
        auto_pick_perks: bool,
        game_mode: int,
        perk_progression_enabled: bool,
        game_tune_started: bool = False,
        rng_marks: dict[str, int] | None = None,
    ) -> WorldEvents:
        def _mark(name: str) -> None:
            if rng_marks is None:
                return
            rng_marks[str(name)] = int(self.state.rng.state)

        dt = float(dt)
        for step in _WORLD_DT_STEPS:
            dt = float(step(dt=dt, players=self.players))
        _mark("ws_begin")
        inputs = normalize_input_frame(inputs, player_count=len(self.players)).as_list()
        prev_positions = [(player.pos.x, player.pos.y) for player in self.players]
        prev_health = [float(player.health) for player in self.players]
        # Native runs `perks_update_effects` early and reads current aim, so stage aim before `player_update`.
        for idx, player in enumerate(self.players):
            input_state = inputs[idx] if idx < len(inputs) else PlayerInput()
            player.aim = input_state.aim
        perks_update_effects(self.state, self.players, dt, creatures=self.creatures.entries, fx_queue=fx_queue)
        _mark("ws_after_perk_effects")
        # `effects_update` runs early in the native frame loop, before creature/projectile updates.
        self.state.effects.update(dt, fx_queue=fx_queue)
        _mark("ws_after_effects_update")
        def _apply_projectile_damage_to_player(player_index: int, damage: float) -> None:
            idx = int(player_index)
            if not (0 <= idx < len(self.players)):
                return
            player_take_projectile_damage(self.state, self.players[idx], float(damage))
        creature_result = self.creatures.update(
            dt,
            dt_ms_i32=(int(dt_ms_i32) if dt_ms_i32 is not None else None),
            state=self.state,
            players=self.players,
            detail_preset=detail_preset,
            world_width=float(world_size),
            world_height=float(world_size),
            fx_queue=fx_queue,
            fx_queue_rotated=fx_queue_rotated,
        )
        _mark("ws_after_creatures")
        deaths = list(creature_result.deaths)
        planned_death_sfx: list[str] = []
        planned_death_sfx_cap = 5
        def _plan_death_sfx_now(death: CreatureDeath) -> None:
            keys = plan_death_sfx_keys([death], rand=self.state.rng.rand)
            if not keys:
                return
            remain = int(planned_death_sfx_cap) - len(planned_death_sfx)
            if remain <= 0:
                return
            planned_death_sfx.extend(keys[:remain])
        for death in deaths:
            _plan_death_sfx_now(death)
        trigger_game_tune = False
        hit_sfx: list[str] = []
        hit_audio_game_tune_started = bool(game_tune_started)
        def _apply_projectile_damage_to_creature(
            creature_index: int,
            damage: float,
            damage_type: int,
            impulse: Vec2,
            owner_id: int,
        ) -> None:
            idx = int(creature_index)
            if not (0 <= idx < len(self.creatures.entries)):
                return
            creature = self.creatures.entries[idx]
            if not creature.active:
                return
            death_start_needed = creature.hp > 0.0 and creature.hitbox_size == CREATURE_HITBOX_ALIVE
            killed = creature_apply_damage(
                creature,
                damage_amount=float(damage),
                damage_type=int(damage_type),
                impulse=impulse,
                owner_id=int(owner_id),
                dt=float(dt),
                players=self.players,
                rand=self.state.rng.rand,
            )
            if killed and death_start_needed:
                self._record_creature_death(
                    creature_index=idx,
                    dt=float(dt),
                    detail_preset=int(detail_preset),
                    world_size=float(world_size),
                    fx_queue=fx_queue,
                    deaths=deaths,
                    plan_death_sfx_now=_plan_death_sfx_now,
                )
        def _on_secondary_detonation_kill(creature_index: int) -> None:
            idx = int(creature_index)
            if not (0 <= idx < len(self.creatures.entries)) or float(self.creatures.entries[idx].hp) > 0.0:
                return
            # Native detonation follow-up re-enters creature death handling but does
            # not run a second death-SFX random pick (`creature_apply_damage` only
            # does that on the original killing hit).
            self._record_creature_death(
                creature_index=idx,
                dt=float(dt),
                detail_preset=int(detail_preset),
                world_size=float(world_size),
                fx_queue=fx_queue,
                deaths=deaths,
                plan_death_sfx_now=_plan_death_sfx_now,
                plan_death_sfx=False,
            )
        def _on_projectile_hit_pre(hit: ProjectileHit) -> ProjectileDecalPostCtx:
            return self._prepare_projectile_hit_presentation(
                hit=hit,
                fx_queue=fx_queue,
                detail_preset=int(detail_preset),
                fx_toggle=int(fx_toggle),
            )
        def _on_projectile_hit_post(_hit: ProjectileHit, post_ctx: object | None) -> None:
            nonlocal trigger_game_tune, hit_audio_game_tune_started
            if not isinstance(post_ctx, ProjectileDecalPostCtx):
                return
            self._finalize_projectile_hit_presentation(
                post_ctx=post_ctx,
                fx_queue=fx_queue,
            )
            if float(self.state.bonuses.freeze) > 0.0:
                if (
                    (not bool(self.state.demo_mode_active))
                    and int(game_mode) != int(GameMode.RUSH)
                    and (not bool(hit_audio_game_tune_started))
                ):
                    trigger_game_tune = True
                    hit_audio_game_tune_started = True
            else:
                hit_trigger, keys = plan_hit_sfx_keys(
                    [_hit],
                    game_mode=int(game_mode),
                    demo_mode_active=bool(self.state.demo_mode_active),
                    game_tune_started=bool(hit_audio_game_tune_started),
                    rand=self.state.rng.rand,
                )
                if hit_trigger:
                    trigger_game_tune = True
                    hit_audio_game_tune_started = True
                if keys:
                    hit_sfx.extend(keys)
        hits = self.state.projectiles.update(
            dt,
            self.creatures.entries,
            world_size=float(world_size),
            damage_scale_by_type=damage_scale_by_type,
            detail_preset=int(detail_preset),
            rng=self.state.rng.rand,
            runtime_state=self.state,
            players=self.players,
            apply_player_damage=_apply_projectile_damage_to_player,
            apply_creature_damage=_apply_projectile_damage_to_creature,
            on_hit=_on_projectile_hit_pre,
            on_hit_post=_on_projectile_hit_post,
        )
        _mark("ws_after_projectiles")
        _mark("ws_after_hit_sfx")
        self.state.secondary_projectiles.update_pulse_gun(
            dt,
            self.creatures.entries,
            apply_creature_damage=_apply_projectile_damage_to_creature,
            runtime_state=self.state,
            fx_queue=fx_queue,
            detail_preset=int(detail_preset),
            on_detonation_kill=_on_secondary_detonation_kill,
        )
        _mark("ws_after_secondary_projectiles")
        for idx, player in enumerate(self.players):
            if idx >= len(prev_health):
                continue
            if float(prev_health[idx]) < 0.0:
                continue
            if float(player.health) >= 0.0:
                continue
            for hook in _PLAYER_DEATH_HOOKS:
                hook(
                    state=self.state,
                    creatures=self.creatures,
                    players=self.players,
                    player=player,
                    dt=float(dt),
                    world_size=float(world_size),
                    detail_preset=int(detail_preset),
                    fx_queue=fx_queue,
                    deaths=deaths,
                )
        def _kill_creature_no_corpse(creature_index: int, owner_id: int) -> None:
            idx = int(creature_index)
            if not (0 <= idx < len(self.creatures.entries)):
                return
            creature = self.creatures.entries[idx]
            if not creature.active:
                return
            if float(creature.hp) <= 0.0:
                return

            creature.last_hit_owner_id = int(owner_id)
            self._record_creature_death(
                creature_index=idx,
                dt=float(dt),
                detail_preset=int(detail_preset),
                world_size=float(world_size),
                fx_queue=fx_queue,
                deaths=deaths,
                plan_death_sfx_now=_plan_death_sfx_now,
                keep_corpse=False,
            )

        self.state.particles.update(
            dt,
            creatures=self.creatures.entries,
            apply_creature_damage=_apply_projectile_damage_to_creature,
            kill_creature=_kill_creature_no_corpse,
            fx_queue=fx_queue,
            sprite_effects=self.state.sprite_effects,
        )
        _mark("ws_after_particles_update")
        self.state.sprite_effects.update(dt)
        _mark("ws_after_sprite_effects")
        _mark("ws_after_particles")
        _mark("ws_after_death_sfx")

        for idx, player in enumerate(self.players):
            input_state = inputs[idx] if idx < len(inputs) else PlayerInput()
            player_update(
                player,
                input_state,
                dt,
                self.state,
                detail_preset=int(detail_preset),
                world_size=float(world_size),
                players=self.players,
            )
            if idx == 0:
                _mark("ws_after_player_update_p0")
        _mark("ws_after_player_update")
        if dt > 0.0:
            self._advance_creature_anim(dt)
            self._advance_player_anim(dt, prev_positions)
        if mid_step_hook is not None:
            mid_step_hook()
        if not bool(defer_camera_shake_update):
            camera_shake_update(self.state, dt)
        _mark("ws_after_camera_update")
        # Native latches `time_scale_active` late (post mode update, pre bonus decrement); next-frame dt uses it.
        self.state.time_scale_active = float(self.state.bonuses.reflex_boost) > 0.0
        bonus_update_pre_pickup_timers(self.state, dt)
        pickups = bonus_update(
            self.state,
            self.players,
            dt,
            creatures=self.creatures.entries,
            update_hud=True,
            apply_creature_damage=_apply_projectile_damage_to_creature,
            detail_preset=int(detail_preset),
            defer_freeze_corpse_fx=bool(defer_freeze_corpse_fx),
        )
        if pickups:
            emit_bonus_pickup_effects(
                state=self.state,
                pickups=pickups,
                detail_preset=int(detail_preset),
            )
        _mark("ws_after_bonus_update")
        if perk_progression_enabled:
            survival_progression_update(
                self.state,
                self.players,
                game_mode=game_mode,
                auto_pick=auto_pick_perks,
                dt=dt,
                creatures=self.creatures.entries,
            )
        _mark("ws_after_progression")
        sfx = list(planned_death_sfx)
        sfx.extend(creature_result.sfx)
        if self.state.sfx_queue:
            sfx.extend(self.state.sfx_queue)
            self.state.sfx_queue.clear()
        _mark("ws_after_sfx_queue_merge")
        # Player-damage VO RNG work lives inside `player_take_damage` for native
        # ordering parity (VO draw before heading-jitter draw).
        _mark("ws_after_player_damage_sfx")
        _mark("ws_after_sfx")

        return WorldEvents(
            hits=hits,
            deaths=tuple(deaths),
            pickups=pickups,
            sfx=sfx,
            trigger_game_tune=bool(trigger_game_tune),
            hit_sfx=hit_sfx,
            death_sfx_preplanned=True,
        )

    def _record_creature_death(
        self,
        *,
        creature_index: int,
        dt: float,
        detail_preset: int,
        world_size: float,
        fx_queue: FxQueue,
        deaths: list[CreatureDeath],
        plan_death_sfx_now: Callable[[CreatureDeath], None],
        keep_corpse: bool = True,
        plan_death_sfx: bool = True,
    ) -> None:
        death = self.creatures.handle_death(
            int(creature_index),
            state=self.state,
            players=self.players,
            rand=self.state.rng.rand,
            dt=float(dt),
            detail_preset=int(detail_preset),
            world_width=float(world_size),
            world_height=float(world_size),
            fx_queue=fx_queue,
            keep_corpse=bool(keep_corpse),
        )
        deaths.append(death)
        if bool(plan_death_sfx):
            plan_death_sfx_now(death)

    def _prepare_projectile_hit_presentation(
        self,
        hit: ProjectileHit,
        *,
        fx_queue: FxQueue,
        detail_preset: int,
        fx_toggle: int,
    ) -> ProjectileDecalPostCtx:
        return queue_projectile_decals_pre_hit(
            state=self.state,
            players=self.players,
            fx_queue=fx_queue,
            hit=hit,
            rand=self.state.rng.rand,
            detail_preset=int(detail_preset),
            fx_toggle=int(fx_toggle),
        )

    def _finalize_projectile_hit_presentation(
        self,
        *,
        post_ctx: ProjectileDecalPostCtx,
        fx_queue: FxQueue,
    ) -> None:
        queue_projectile_decals_post_hit(
            fx_queue=fx_queue,
            post_ctx=post_ctx,
            rand=self.state.rng.rand,
        )

    def _advance_creature_anim(self, dt: float) -> None:
        if float(self.state.bonuses.freeze) > 0.0:
            return
        for creature in self.creatures.entries:
            if not (creature.active and creature.hp > 0.0):
                continue
            try:
                type_id = CreatureTypeId(int(creature.type_id))
            except ValueError:
                continue
            info = CREATURE_ANIM.get(type_id)
            if info is None:
                continue
            creature.anim_phase, _ = creature_anim_advance_phase(
                creature.anim_phase,
                anim_rate=info.anim_rate,
                move_speed=float(creature.move_speed),
                dt=dt,
                size=float(creature.size),
                local_scale=float(getattr(creature, "move_scale", 1.0)),
                flags=creature.flags,
                ai_mode=int(creature.ai_mode),
            )

    def _advance_player_anim(self, dt: float, prev_positions: list[tuple[float, float]]) -> None:
        info = CREATURE_ANIM.get(CreatureTypeId.TROOPER)
        if info is None:
            return
        for idx, player in enumerate(self.players):
            if idx >= len(prev_positions):
                continue
            prev_x, prev_y = prev_positions[idx]
            speed = Vec2(player.pos.x - prev_x, player.pos.y - prev_y).length()
            move_speed = speed / dt / 120.0 if dt > 0.0 else 0.0
            player.move_phase, _ = creature_anim_advance_phase(
                player.move_phase,
                anim_rate=info.anim_rate,
                move_speed=move_speed,
                dt=dt,
                size=float(player.size),
                local_scale=1.0,
                flags=CreatureFlags(0),
                ai_mode=0,
            )



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
import datetime as dt
import hashlib
import random

import pyray as rl

from grim.assets import PaqTextureCache
from grim.audio import AudioState
from grim.console import ConsoleState
from grim.config import CrimsonConfig
from grim.geom import Rect, Vec2
from grim.math import clamp
from grim.view import ViewContext

from ..debug import debug_enabled
from ..game_modes import GameMode
from ..gameplay import (
    survival_check_level_up,
    weapon_assign_player,
)
from ..ui.cursor import draw_aim_cursor, draw_menu_cursor
from ..ui.hud import draw_hud_overlay, hud_flags_for_game_mode
from ..input_codes import (
    config_keybinds_for_player,
    input_code_is_down_for_player,
    input_code_is_pressed_for_player,
)
from ..ui.perk_menu import PERK_MENU_TRANSITION_MS, draw_ui_text, load_perk_menu_assets
from ..weapons import WEAPON_BY_ID
from ..replay import ReplayHeader, ReplayRecorder, ReplayStatusSnapshot, dump_replay
from ..replay.types import WEAPON_USAGE_COUNT
from ..replay.checkpoints import (
    FORMAT_VERSION as CHECKPOINTS_FORMAT_VERSION,
    ReplayCheckpoint,
    ReplayCheckpoints,
    build_checkpoint,
    default_checkpoints_path,
    dump_checkpoints_file,
    resolve_checkpoint_sample_rate,
)
from ..sim.clock import FixedStepClock
from ..sim.sessions import SurvivalDeterministicSession
from .base_gameplay_mode import BaseGameplayMode
from .components.highscore_record_builder import build_highscore_record_for_game_over
from .components.perk_menu_controller import PerkMenuContext, PerkMenuController

WORLD_SIZE = 1024.0

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_SPONSOR_COLOR = rl.Color(255, 255, 255, int(255 * 0.5))
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)

PERK_PROMPT_MAX_TIMER_MS = 200.0
PERK_PROMPT_OUTSET_X = 50.0
# Perk prompt bar geometry comes from `ui_menu_assets_init` + `ui_menu_layout_init`:
# - `ui_menu_item_element` is set_rect(512x64, offset -72,-60)
# - the perk prompt mutates quad coords: x = (x - 300) * 0.75, y = y * 0.75
PERK_PROMPT_BAR_SCALE = 0.75
PERK_PROMPT_BAR_BASE_OFFSET_X = -72.0
PERK_PROMPT_BAR_BASE_OFFSET_Y = -60.0
PERK_PROMPT_BAR_SHIFT_X = -300.0

# `ui_textLevelUp` is set_rect(75x25, offset -230,-27), then its quad coords are:
# x = x * 0.85 - 46, y = y * 0.85 - 4
PERK_PROMPT_LEVEL_UP_SCALE = 0.85
PERK_PROMPT_LEVEL_UP_BASE_OFFSET_X = -230.0
PERK_PROMPT_LEVEL_UP_BASE_OFFSET_Y = -27.0
PERK_PROMPT_LEVEL_UP_BASE_W = 75.0
PERK_PROMPT_LEVEL_UP_BASE_H = 25.0
PERK_PROMPT_LEVEL_UP_SHIFT_X = -46.0
PERK_PROMPT_LEVEL_UP_SHIFT_Y = -4.0

PERK_PROMPT_TEXT_MARGIN_X = 16.0
PERK_PROMPT_TEXT_OFFSET_Y = 8.0

_DEBUG_WEAPON_IDS = tuple(sorted(WEAPON_BY_ID))


@dataclass(slots=True)
class _SurvivalState:
    elapsed_ms: float = 0.0
    stage: int = 0
    spawn_cooldown: float = 0.0


class SurvivalMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        texture_cache: PaqTextureCache | None = None,
        config: CrimsonConfig | None = None,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: random.Random | None = None,
    ) -> None:
        super().__init__(
            ctx,
            world_size=WORLD_SIZE,
            default_game_mode_id=int(GameMode.SURVIVAL),
            demo_mode_active=False,
            difficulty_level=0,
            hardcore=False,
            texture_cache=texture_cache,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._survival = _SurvivalState()

        self._perk_prompt_timer_ms = 0.0
        self._perk_prompt_hover = False
        self._perk_prompt_pulse = 0.0
        self._perk_menu = PerkMenuController(on_close=self._reset_perk_prompt, on_pick=self._record_perk_pick)
        self._hud_fade_ms = PERK_MENU_TRANSITION_MS
        self._perk_menu_assets = None
        self._cursor_time = 0.0
        self._sim_clock = FixedStepClock(tick_rate=60)
        self._replay_recorder: ReplayRecorder | None = None
        self._replay_checkpoints: list[ReplayCheckpoint] = []
        self._replay_checkpoints_sample_rate: int = 60
        self._replay_checkpoints_last_tick: int | None = None
        self._sim_session: SurvivalDeterministicSession | None = None

    def _reset_perk_prompt(self) -> None:
        if int(self._state.perk_selection.pending_count) > 0:
            # Reset the prompt swing so each pending perk replays the intro.
            self._perk_prompt_timer_ms = 0.0
            self._perk_prompt_hover = False
            self._perk_prompt_pulse = 0.0

    def _record_perk_pick(self, choice_index: int) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        recorder.record_perk_pick(player_index=0, choice_index=int(choice_index))

    def _record_replay_checkpoint(
        self,
        tick_index: int,
        *,
        force: bool = False,
        rng_marks: dict[str, int] | None = None,
        deaths: Sequence[object] | None = None,
        events: object | None = None,
        command_hash: str | None = None,
    ) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        if tick_index < 0:
            return
        if not force and (tick_index % int(self._replay_checkpoints_sample_rate or 1)) != 0:
            return
        if self._replay_checkpoints_last_tick == int(tick_index):
            return
        self._replay_checkpoints.append(
            build_checkpoint(
                tick_index=int(tick_index),
                world=self._world.world_state,
                elapsed_ms=float(self._survival.elapsed_ms),
                rng_marks=rng_marks,
                deaths=deaths,
                events=events,
                command_hash=command_hash,
            )
        )
        self._replay_checkpoints_last_tick = int(tick_index)

    def _save_replay(self) -> None:
        recorder = self._replay_recorder
        if recorder is None:
            return
        replay = recorder.finish()
        self._record_replay_checkpoint(max(0, recorder.tick_index - 1), force=True)
        terminal_tick = int(recorder.tick_index)
        if any(int(getattr(event, "tick_index", -1)) == terminal_tick for event in replay.events):
            self._record_replay_checkpoint(terminal_tick, force=True)
        data = dump_replay(replay)
        digest = hashlib.sha256(data).hexdigest()
        stamp = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
        replay_dir = self._base_dir / "replays"
        replay_dir.mkdir(parents=True, exist_ok=True)
        score = int(self._player.experience)
        base_name = f"survival_{stamp}_score{score}"
        path = replay_dir / f"{base_name}.crdemo.gz"
        counter = 1
        while path.exists():
            path = replay_dir / f"{base_name}_{counter}.crdemo.gz"
            counter += 1
        path.write_bytes(data)
        checkpoints_path = default_checkpoints_path(path)
        dump_checkpoints_file(
            checkpoints_path,
            ReplayCheckpoints(
                version=CHECKPOINTS_FORMAT_VERSION,
                replay_sha256=digest,
                sample_rate=int(self._replay_checkpoints_sample_rate or 0),
                checkpoints=list(self._replay_checkpoints),
            ),
        )
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        if self._console is not None:
            self._console.log.log(f"replay: saved {path}")
            self._console.log.log(f"replay: saved {checkpoints_path}")
            self._console.log.flush()

    def _perk_menu_context(self) -> PerkMenuContext:
        fx_toggle = int(self._config.data.get("fx_toggle", 0) or 0) if self._config is not None else 0
        fx_detail = bool(int(self._config.data.get("fx_detail_0", 0) or 0)) if self._config is not None else False
        players = self._world.players
        return PerkMenuContext(
            state=self._state,
            perk_state=self._state.perk_selection,
            players=players,
            creatures=self._creatures.entries,
            player=self._player,
            game_mode=int(GameMode.SURVIVAL),
            player_count=len(players),
            fx_toggle=fx_toggle,
            fx_detail=fx_detail,
            font=self._small,
            assets=self._perk_menu_assets,
            mouse=self._ui_mouse_pos(),
            play_sfx=self._world.audio_router.play_sfx,
        )

    def _wrap_ui_text(self, text: str, *, max_width: float, scale: float = UI_TEXT_SCALE) -> list[str]:
        lines: list[str] = []
        for raw in text.splitlines() or [""]:
            para = raw.strip()
            if not para:
                lines.append("")
                continue
            current = ""
            for word in para.split():
                candidate = word if not current else f"{current} {word}"
                if current and self._ui_text_width(candidate, scale) > max_width:
                    lines.append(current)
                    current = word
                else:
                    current = candidate
            if current:
                lines.append(current)
        return lines

    def _camera_world_to_screen(self, pos: Vec2) -> Vec2:
        return self._world.world_to_screen(pos)

    def _camera_screen_to_world(self, pos: Vec2) -> Vec2:
        return self._world.screen_to_world(pos)

    def open(self) -> None:
        super().open()

        self._perk_menu_assets = load_perk_menu_assets(self._assets_root)
        if self._perk_menu_assets.missing:
            self._missing_assets.extend(self._perk_menu_assets.missing)
        self._perk_menu.reset()
        self._cursor_time = 0.0
        self._cursor_pulse_time = 0.0
        self._sim_clock.reset()
        self._survival = _SurvivalState()
        self._sim_session = SurvivalDeterministicSession(
            world=self._world.world_state,
            world_size=float(self._world.world_size),
            damage_scale_by_type=self._world._damage_scale_by_type,
            fx_queue=self._world.fx_queue,
            fx_queue_rotated=self._world.fx_queue_rotated,
            detail_preset=5,
            fx_toggle=0,
            game_tune_started=bool(self._world._game_tune_started),
            clear_fx_queues_each_tick=False,
        )

        self._perk_prompt_timer_ms = 0.0
        self._perk_prompt_hover = False
        self._perk_prompt_pulse = 0.0
        self._hud_fade_ms = PERK_MENU_TRANSITION_MS
        status = self._state.status
        weapon_usage_counts: tuple[int, ...] = ()
        if status is not None:
            try:
                raw_counts = status.data.get("weapon_usage_counts")
                if isinstance(raw_counts, list):
                    weapon_usage_counts = tuple(int(value) & 0xFFFFFFFF for value in raw_counts[:WEAPON_USAGE_COUNT])
            except Exception:
                weapon_usage_counts = ()
        if len(weapon_usage_counts) != WEAPON_USAGE_COUNT:
            weapon_usage_counts = tuple(weapon_usage_counts) + (0,) * max(
                0, WEAPON_USAGE_COUNT - len(weapon_usage_counts)
            )
            weapon_usage_counts = weapon_usage_counts[:WEAPON_USAGE_COUNT]
        status_snapshot = ReplayStatusSnapshot(
            quest_unlock_index=int(getattr(status, "quest_unlock_index", 0) or 0) if status is not None else 0,
            quest_unlock_index_full=int(getattr(status, "quest_unlock_index_full", 0) or 0)
            if status is not None
            else 0,
            weapon_usage_counts=weapon_usage_counts,
        )
        self._replay_recorder = ReplayRecorder(
            ReplayHeader(
                game_mode_id=int(GameMode.SURVIVAL),
                seed=int(self._state.rng.state),
                tick_rate=int(self._sim_clock.tick_rate),
                difficulty_level=int(self._world.difficulty_level),
                hardcore=bool(self._world.hardcore),
                preserve_bugs=bool(self._world.preserve_bugs),
                world_size=float(self._world.world_size),
                player_count=len(self._world.players),
                status=status_snapshot,
            )
        )
        tick_rate = int(self._replay_recorder.header.tick_rate)
        self._replay_checkpoints_sample_rate = resolve_checkpoint_sample_rate(tick_rate)
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

    def close(self) -> None:
        if self._perk_menu_assets is not None:
            self._perk_menu_assets = None
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        self._sim_session = None
        super().close()

    def _handle_input(self) -> None:
        if self._game_over_active:
            if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
                self._action = "back_to_menu"
                self.close_requested = True
            return
        if self._perk_menu.open and rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._world.audio_router.play_sfx("sfx_ui_buttonclick")
            self._perk_menu.close()
            return

        if rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if debug_enabled() and (not self._perk_menu.open):
            if rl.is_key_pressed(rl.KeyboardKey.KEY_F2):
                self._state.debug_god_mode = not bool(self._state.debug_god_mode)
                self._world.audio_router.play_sfx("sfx_ui_buttonclick")
            if rl.is_key_pressed(rl.KeyboardKey.KEY_F3):
                self._state.perk_selection.pending_count += 1
                self._state.perk_selection.choices_dirty = True
                self._world.audio_router.play_sfx("sfx_ui_levelup")
            if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT_BRACKET):
                self._debug_cycle_weapon(-1)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT_BRACKET):
                self._debug_cycle_weapon(1)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_X):
                self._player.experience += 5000
                survival_check_level_up(self._player, self._state.perk_selection)

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._action = "open_pause_menu"
            return

    def _debug_cycle_weapon(self, delta: int) -> None:
        weapon_ids = _DEBUG_WEAPON_IDS
        if not weapon_ids:
            return
        current = int(self._player.weapon_id)
        try:
            idx = weapon_ids.index(current)
        except ValueError:
            idx = 0
        weapon_id = int(weapon_ids[(idx + int(delta)) % len(weapon_ids)])
        weapon_assign_player(self._player, weapon_id, state=self._state)

    def _player_name_default(self) -> str:
        config = self._config
        if config is None:
            return ""
        raw = config.data.get("player_name")
        if isinstance(raw, (bytes, bytearray)):
            return bytes(raw).split(b"\x00", 1)[0].decode("latin-1", errors="ignore")
        if isinstance(raw, str):
            return raw
        return ""

    def _enter_game_over(self) -> None:
        if self._game_over_active:
            return
        game_mode_id = int(self._config.data.get("game_mode", 1)) if self._config is not None else 1
        record = build_highscore_record_for_game_over(
            state=self._state,
            player=self._player,
            survival_elapsed_ms=int(self._survival.elapsed_ms),
            creature_kill_count=int(self._creatures.kill_count),
            game_mode_id=game_mode_id,
        )
        self._game_over_record = record
        self._game_over_ui.open()
        self._game_over_active = True
        self._perk_menu.close()
        self._save_replay()

    def _perk_prompt_label(self) -> str:
        if self._config is not None and not bool(int(self._config.data.get("ui_info_texts", 1) or 0)):
            return ""
        pending = int(self._state.perk_selection.pending_count)
        if pending <= 0:
            return ""
        suffix = f" ({pending})" if pending > 1 else ""
        return f"Press Mouse2 to pick a perk{suffix}"

    def _perk_prompt_hinge(self) -> Vec2:
        screen_w = float(rl.get_screen_width())
        hinge_x = screen_w + PERK_PROMPT_OUTSET_X
        hinge_y = 80.0 if int(screen_w) == 640 else 40.0
        return Vec2(hinge_x, hinge_y)

    def _perk_prompt_rect(self, label: str, *, scale: float = UI_TEXT_SCALE) -> Rect:
        hinge = self._perk_prompt_hinge()
        if self._perk_menu_assets is not None and self._perk_menu_assets.menu_item is not None:
            tex = self._perk_menu_assets.menu_item
            bar_w = float(tex.width) * PERK_PROMPT_BAR_SCALE
            bar_h = float(tex.height) * PERK_PROMPT_BAR_SCALE
            local_x = (PERK_PROMPT_BAR_BASE_OFFSET_X + PERK_PROMPT_BAR_SHIFT_X) * PERK_PROMPT_BAR_SCALE
            local_y = PERK_PROMPT_BAR_BASE_OFFSET_Y * PERK_PROMPT_BAR_SCALE

            return Rect.from_top_left(
                hinge.offset(dx=local_x, dy=local_y),
                bar_w,
                bar_h,
            )

        margin = 16.0 * scale
        text_w = float(self._ui_text_width(label, scale))
        text_h = float(self._ui_line_height(scale))
        x = float(rl.get_screen_width()) - margin - text_w
        y = margin
        return Rect.from_top_left(Vec2(x, y), text_w, text_h)

    def update(self, dt: float) -> None:
        self._update_audio(dt)

        dt_frame, dt_ui_ms = self._tick_frame(dt)
        self._cursor_time += dt_frame
        self._handle_input()
        if self._action == "open_pause_menu":
            return

        if self._game_over_active:
            self._update_game_over_ui(dt)
            return

        any_alive = any(player.health > 0.0 for player in self._world.players)
        perk_pending = int(self._state.perk_selection.pending_count) > 0 and any_alive

        self._perk_prompt_hover = False
        perk_ctx = self._perk_menu_context()
        if self._perk_menu.open:
            self._perk_menu.handle_input(perk_ctx, dt_frame=dt_frame, dt_ui_ms=dt_ui_ms)

        perk_menu_active = self._perk_menu.active
        any_alive = self._any_player_alive()

        if (not perk_menu_active) and perk_pending and (not self._paused):
            label = self._perk_prompt_label()
            if label:
                rect = self._perk_prompt_rect(label)
                mouse = self._ui_mouse_pos()
                self._perk_prompt_hover = rect.contains(mouse)

            player0_binds = config_keybinds_for_player(self._config, player_index=0)
            fire_key = 0x100
            if len(player0_binds) >= 5:
                fire_key = int(player0_binds[4])

            pick_key = 0x101
            if self._config is not None:
                pick_key = int(self._config.data.get("keybind_pick_perk", pick_key) or pick_key)

            if input_code_is_pressed_for_player(pick_key, player_index=0) and (
                not input_code_is_down_for_player(fire_key, player_index=0)
            ):
                self._perk_prompt_pulse = 1000.0
                if self._replay_recorder is not None:
                    self._record_replay_checkpoint(max(0, self._replay_recorder.tick_index - 1), force=True)
                opened = self._perk_menu.open_if_available(perk_ctx)
                if opened and self._replay_recorder is not None:
                    self._replay_recorder.record_perk_menu_open(player_index=0)
            elif self._perk_prompt_hover and input_code_is_pressed_for_player(fire_key, player_index=0):
                self._perk_prompt_pulse = 1000.0
                if self._replay_recorder is not None:
                    self._record_replay_checkpoint(max(0, self._replay_recorder.tick_index - 1), force=True)
                opened = self._perk_menu.open_if_available(perk_ctx)
                if opened and self._replay_recorder is not None:
                    self._replay_recorder.record_perk_menu_open(player_index=0)

        if not self._paused and not self._game_over_active:
            pulse_delta = dt_ui_ms * (6.0 if self._perk_prompt_hover else -2.0)
            self._perk_prompt_pulse = clamp(self._perk_prompt_pulse + pulse_delta, 0.0, 1000.0)

        sim_active = (not self._paused) and any_alive and (not perk_menu_active)

        prompt_active = perk_pending and (not perk_menu_active) and (not self._paused)
        if prompt_active:
            self._perk_prompt_timer_ms = clamp(self._perk_prompt_timer_ms + dt_ui_ms, 0.0, PERK_PROMPT_MAX_TIMER_MS)
        else:
            self._perk_prompt_timer_ms = clamp(self._perk_prompt_timer_ms - dt_ui_ms, 0.0, PERK_PROMPT_MAX_TIMER_MS)

        self._perk_menu.tick_timeline(dt_ui_ms)
        if self._perk_menu.active:
            self._hud_fade_ms = 0.0
        else:
            self._hud_fade_ms = clamp(self._hud_fade_ms + dt_ui_ms, 0.0, PERK_MENU_TRANSITION_MS)

        if not sim_active:
            self._sim_clock.reset()
            if not any_alive:
                self._enter_game_over()
            return

        ticks_to_run = self._sim_clock.advance(dt_frame)
        if ticks_to_run <= 0:
            return

        dt_tick = float(self._sim_clock.dt_tick)
        input_frame = self._build_local_inputs(dt_frame=dt_frame)
        session = self._sim_session
        if session is None:
            return
        if self._world.audio_router is not None:
            self._world.audio_router.audio = self._world.audio
            self._world.audio_router.audio_rng = self._world.audio_rng
            self._world.audio_router.demo_mode_active = self._world.demo_mode_active
        if self._world.ground is not None:
            self._world._sync_ground_settings()
            self._world.ground.process_pending()
        detail_preset = 5
        fx_toggle = 0
        if self._config is not None:
            detail_preset = int(self._config.data.get("detail_preset", 5) or 5)
            fx_toggle = int(self._config.data.get("fx_toggle", 0) or 0)
        session.detail_preset = int(detail_preset)
        session.fx_toggle = int(fx_toggle)

        for tick_offset in range(int(ticks_to_run)):
            inputs = input_frame if tick_offset == 0 else self._clear_local_input_edges(input_frame)
            recorder = self._replay_recorder
            if recorder is not None:
                tick_index = recorder.record_tick(inputs)
            else:
                tick_index = None
            tick = session.step_tick(
                dt_frame=dt_tick,
                inputs=inputs,
            )
            self._world.apply_step_result(
                tick.step,
                game_tune_started=bool(session.game_tune_started),
                apply_audio=True,
                update_camera=True,
            )
            self._survival.elapsed_ms = float(session.elapsed_ms)
            self._survival.stage = int(session.stage)
            self._survival.spawn_cooldown = float(session.spawn_cooldown_ms)
            world_events = tick.step.events

            if tick_index is not None:
                self._record_replay_checkpoint(
                    int(tick_index),
                    rng_marks=tick.rng_marks,
                    deaths=world_events.deaths,
                    events=world_events,
                    command_hash=str(tick.step.command_hash),
                )

            if not any(player.health > 0.0 for player in self._world.players):
                self._enter_game_over()
                break

    def _draw_perk_prompt(self) -> None:
        if self._game_over_active:
            return
        if self._perk_menu.active:
            return
        if not any(player.health > 0.0 for player in self._world.players):
            return
        pending = int(self._state.perk_selection.pending_count)
        if pending <= 0:
            return
        label = self._perk_prompt_label()
        if not label:
            return

        alpha = float(self._perk_prompt_timer_ms) / PERK_PROMPT_MAX_TIMER_MS
        if alpha <= 1e-3:
            return

        hinge = self._perk_prompt_hinge()
        # Prompt swings counter-clockwise; raylib's Y-down makes positive rotation clockwise.
        rot_deg = -(1.0 - alpha) * 90.0
        tint = rl.Color(255, 255, 255, int(255 * alpha))

        text_w = float(self._ui_text_width(label, UI_TEXT_SCALE))
        x = float(rl.get_screen_width()) - PERK_PROMPT_TEXT_MARGIN_X - text_w
        y = hinge.y + PERK_PROMPT_TEXT_OFFSET_Y
        color = rl.Color(UI_TEXT_COLOR.r, UI_TEXT_COLOR.g, UI_TEXT_COLOR.b, int(255 * alpha))
        draw_ui_text(self._small, label, Vec2(x, y), scale=UI_TEXT_SCALE, color=color)

        if self._perk_menu_assets is not None and self._perk_menu_assets.menu_item is not None:
            tex = self._perk_menu_assets.menu_item
            bar_w = float(tex.width) * PERK_PROMPT_BAR_SCALE
            bar_h = float(tex.height) * PERK_PROMPT_BAR_SCALE
            local_x = (PERK_PROMPT_BAR_BASE_OFFSET_X + PERK_PROMPT_BAR_SHIFT_X) * PERK_PROMPT_BAR_SCALE
            local_y = PERK_PROMPT_BAR_BASE_OFFSET_Y * PERK_PROMPT_BAR_SCALE
            src = rl.Rectangle(float(tex.width), 0.0, -float(tex.width), float(tex.height))
            dst = rl.Rectangle(hinge.x, hinge.y, bar_w, bar_h)
            origin = rl.Vector2(float(-local_x), float(-local_y))
            rl.draw_texture_pro(tex, src, dst, origin, rot_deg, tint)

        if self._perk_menu_assets is not None and self._perk_menu_assets.title_level_up is not None:
            tex = self._perk_menu_assets.title_level_up
            local_x = PERK_PROMPT_LEVEL_UP_BASE_OFFSET_X * PERK_PROMPT_LEVEL_UP_SCALE + PERK_PROMPT_LEVEL_UP_SHIFT_X
            local_y = PERK_PROMPT_LEVEL_UP_BASE_OFFSET_Y * PERK_PROMPT_LEVEL_UP_SCALE + PERK_PROMPT_LEVEL_UP_SHIFT_Y
            w = PERK_PROMPT_LEVEL_UP_BASE_W * PERK_PROMPT_LEVEL_UP_SCALE
            h = PERK_PROMPT_LEVEL_UP_BASE_H * PERK_PROMPT_LEVEL_UP_SCALE
            pulse_alpha = (100.0 + float(int(self._perk_prompt_pulse * 155.0 / 1000.0))) / 255.0
            pulse_alpha = max(0.0, min(1.0, pulse_alpha))
            label_alpha = max(0.0, min(1.0, alpha * pulse_alpha))
            pulse_tint = rl.Color(255, 255, 255, int(255 * label_alpha))
            src = rl.Rectangle(0.0, 0.0, float(tex.width), float(tex.height))
            dst = rl.Rectangle(hinge.x, hinge.y, w, h)
            origin = rl.Vector2(float(-local_x), float(-local_y))
            rl.draw_texture_pro(tex, src, dst, origin, rot_deg, pulse_tint)
            if label_alpha > 0.0:
                rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
                rl.draw_texture_pro(tex, src, dst, origin, rot_deg, pulse_tint)
                rl.end_blend_mode()

    def _draw_game_cursor(self) -> None:
        mouse_pos = self._ui_mouse
        cursor_tex = self._perk_menu_assets.cursor if self._perk_menu_assets is not None else None
        draw_menu_cursor(
            self._world.particles_texture,
            cursor_tex,
            pos=mouse_pos,
            pulse_time=float(self._cursor_pulse_time),
        )

    def _draw_aim_cursor(self) -> None:
        mouse_pos = self._ui_mouse
        aim_tex = self._perk_menu_assets.aim if self._perk_menu_assets is not None else None
        draw_aim_cursor(self._world.particles_texture, aim_tex, pos=mouse_pos)

    def draw(self) -> None:
        perk_menu_active = self._perk_menu.active
        self._world.draw(draw_aim_indicators=(not self._game_over_active) and (not perk_menu_active))
        self._draw_screen_fade()

        hud_bottom = 0.0
        if (not self._game_over_active) and (not perk_menu_active) and self._hud_assets is not None:
            hud_alpha = clamp(self._hud_fade_ms / PERK_MENU_TRANSITION_MS, 0.0, 1.0)
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar(alpha=hud_alpha)
            hud_bottom = draw_hud_overlay(
                self._hud_assets,
                state=self._hud_state,
                player=self._player,
                players=self._world.players,
                bonus_hud=self._state.bonus_hud,
                elapsed_ms=self._survival.elapsed_ms,
                score=self._player.experience,
                font=self._small,
                alpha=hud_alpha,
                frame_dt_ms=self._last_dt_ms,
                show_health=hud_flags.show_health,
                show_weapon=hud_flags.show_weapon,
                show_xp=hud_flags.show_xp,
                show_time=hud_flags.show_time,
                show_quest_hud=hud_flags.show_quest_hud,
                small_indicators=self._hud_small_indicators(),
            )

        if debug_enabled() and (not self._game_over_active) and (not perk_menu_active):
            # Minimal debug text.
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            line = float(self._ui_line_height())
            self._draw_ui_text(
                f"survival: t={self._survival.elapsed_ms / 1000.0:6.1f}s  stage={self._survival.stage}",
                Vec2(x, y),
                UI_TEXT_COLOR,
            )
            self._draw_ui_text(
                f"xp={self._player.experience}  level={self._player.level}  kills={self._creatures.kill_count}",
                Vec2(x, y + line),
                UI_HINT_COLOR,
            )
            god = "on" if self._state.debug_god_mode else "off"
            self._draw_ui_text(
                f"debug: [/] weapon  F3 perk+1  F2 god={god}  X xp+5000",
                Vec2(x, y + line * 2.0),
                UI_HINT_COLOR,
                scale=0.9,
            )
            if self._paused:
                self._draw_ui_text("paused (TAB)", Vec2(x, y + line * 3.0), UI_HINT_COLOR)
            if self._player.health <= 0.0:
                self._draw_ui_text("game over", Vec2(x, y + line * 3.0), UI_ERROR_COLOR)
        warn_y = float(rl.get_screen_height()) - 28.0
        if self._world.missing_assets:
            warn = "Missing world assets: " + ", ".join(self._world.missing_assets)
            self._draw_ui_text(warn, Vec2(24.0, warn_y), UI_ERROR_COLOR, scale=0.8)
            warn_y -= float(self._ui_line_height(scale=0.8)) + 2.0
        if self._hud_missing:
            warn = "Missing HUD assets: " + ", ".join(self._hud_missing)
            self._draw_ui_text(warn, Vec2(24.0, warn_y), UI_ERROR_COLOR, scale=0.8)

        self._draw_perk_prompt()
        if not self._game_over_active:
            self._perk_menu.draw(self._perk_menu_context())
        if (not self._game_over_active) and perk_menu_active:
            self._draw_game_cursor()
        if (not self._game_over_active) and (not perk_menu_active):
            self._draw_aim_cursor()

        if self._game_over_active and self._game_over_record is not None:
            self._game_over_ui.draw(
                record=self._game_over_record,
                banner_kind=self._game_over_banner,
                hud_assets=self._hud_assets,
                mouse=self._ui_mouse_pos(),
            )



from __future__ import annotations

from dataclasses import dataclass, field
import math
from typing import Callable, Protocol, Sequence

from grim.color import RGBA
from grim.geom import Vec2
from grim.math import clamp

from .effects_atlas import EffectId

__all__ = [
    "FX_QUEUE_CAPACITY",
    "FX_QUEUE_MAX_COUNT",
    "FX_QUEUE_ROTATED_CAPACITY",
    "FX_QUEUE_ROTATED_MAX_COUNT",
    "EFFECT_POOL_SIZE",
    "PARTICLE_POOL_SIZE",
    "SPRITE_EFFECT_POOL_SIZE",
    "FxQueue",
    "FxQueueEntry",
    "FxQueueRotated",
    "FxQueueRotatedEntry",
    "EffectEntry",
    "EffectPool",
    "Particle",
    "ParticlePool",
    "SpriteEffect",
    "SpriteEffectPool",
]

EFFECT_POOL_SIZE = 0x200
PARTICLE_POOL_SIZE = 0x80
SPRITE_EFFECT_POOL_SIZE = 0x180

FX_QUEUE_CAPACITY = 0x80
FX_QUEUE_MAX_COUNT = 0x7F

FX_QUEUE_ROTATED_CAPACITY = 0x40
FX_QUEUE_ROTATED_MAX_COUNT = 0x3F


def _default_rand() -> int:
    return 0


class _CreatureForParticles(Protocol):
    active: bool
    pos: Vec2
    hp: float
    size: float
    hitbox_size: float
    tint: RGBA


CreatureDamageApplier = Callable[[int, float, int, Vec2, int], None]
CreatureKillHandler = Callable[[int, int], None]


@dataclass(slots=True)
class Particle:
    active: bool = False
    render_flag: bool = False
    pos: Vec2 = field(default_factory=Vec2)
    vel: Vec2 = field(default_factory=Vec2)
    scale_x: float = 1.0
    scale_y: float = 1.0
    scale_z: float = 1.0
    age: float = 0.0
    intensity: float = 0.0
    angle: float = 0.0
    spin: float = 0.0
    style_id: int = 0
    target_id: int = -1
    owner_id: int = -100


class ParticlePool:
    def __init__(self, *, size: int = PARTICLE_POOL_SIZE, rand: Callable[[], int] | None = None) -> None:
        self._entries = [Particle() for _ in range(int(size))]
        self._rand = rand or _default_rand

    @property
    def entries(self) -> list[Particle]:
        return self._entries

    def reset(self) -> None:
        for entry in self._entries:
            entry.active = False

    def _alloc_slot(self) -> int:
        for i, entry in enumerate(self._entries):
            if not entry.active:
                return i
        if not self._entries:
            raise ValueError("Particle pool has zero entries")
        # Native: `crt_rand() & 0x7f` (pool size is 0x80).
        return int(self._rand()) % len(self._entries)

    def spawn_particle(
        self,
        *,
        pos: Vec2,
        angle: float,
        intensity: float = 1.0,
        owner_id: int = -100,
    ) -> int:
        """Port of `fx_spawn_particle` (0x00420130)."""

        idx = self._alloc_slot()
        entry = self._entries[idx]
        entry.active = True
        entry.render_flag = True
        entry.pos = pos
        entry.vel = Vec2.from_angle(angle) * 90.0
        entry.scale_x = 1.0
        entry.scale_y = 1.0
        entry.scale_z = 1.0
        entry.age = 0.0
        entry.intensity = float(intensity)
        entry.angle = float(angle)
        entry.spin = float(int(self._rand()) % 0x274) * 0.01
        entry.style_id = 0
        entry.target_id = -1
        entry.owner_id = int(owner_id)
        return idx

    def spawn_particle_slow(
        self,
        *,
        pos: Vec2,
        angle: float,
        owner_id: int = -100,
    ) -> int:
        """Port of `fx_spawn_particle_slow` (0x00420240)."""

        idx = self._alloc_slot()
        entry = self._entries[idx]
        entry.active = True
        entry.render_flag = True
        entry.pos = pos
        entry.vel = Vec2.from_angle(angle) * 30.0
        entry.scale_x = 1.0
        entry.scale_y = 1.0
        entry.scale_z = 1.0
        entry.age = 0.0
        entry.intensity = 1.0
        entry.angle = float(angle)
        entry.spin = float(int(self._rand()) % 0x274) * 0.01
        entry.style_id = 8
        entry.target_id = -1
        entry.owner_id = int(owner_id)
        return idx

    def iter_active(self) -> list[Particle]:
        return [entry for entry in self._entries if entry.active]

    def update(
        self,
        dt: float,
        *,
        creatures: Sequence[_CreatureForParticles] | None = None,
        apply_creature_damage: CreatureDamageApplier | None = None,
        kill_creature: CreatureKillHandler | None = None,
        fx_queue: FxQueue | None = None,
        sprite_effects: SpriteEffectPool | None = None,
    ) -> list[int]:
        """Advance particles and deactivate expired entries.

        This is a minimal port of the particle loop inside `projectile_update`
        (0x00420b90). It captures the per-style decay/movement rules that drive
        visual lifetimes and the weapon-driven collision damage.

        Returns indices of particles that were deactivated this tick.
        """

        if dt <= 0.0:
            return []

        def _creature_find_in_radius(*, pos: Vec2, radius: float) -> int:
            if creatures is None:
                return -1
            max_index = min(len(creatures), 0x180)
            radius = float(radius)

            for creature_idx in range(max_index):
                creature = creatures[creature_idx]
                if not creature.active:
                    continue
                if creature.hp <= 0.0:
                    continue
                if creature.hitbox_size < 5.0:
                    continue

                size = float(creature.size)
                dist = creature.pos.distance_to(pos) - radius
                threshold = size * 0.14285715 + 3.0
                if threshold < dist:
                    continue
                return int(creature_idx)

            return -1

        expired: list[int] = []
        rand = self._rand

        for idx, entry in enumerate(self._entries):
            if not entry.active:
                continue

            style = int(entry.style_id) & 0xFF

            if style == 8:
                entry.intensity -= dt * 0.11
                entry.spin += dt * 5.0
                move_scale = entry.intensity
                if move_scale <= 0.15:
                    move_scale *= 0.55
                entry.pos = entry.pos + entry.vel * (dt * move_scale)
            else:
                entry.intensity -= dt * 0.9
                entry.spin += dt
                move_scale = max(entry.intensity, 0.15) * 2.5
                entry.pos = entry.pos + entry.vel * (dt * move_scale)

            if entry.render_flag:
                # Random walk drift (native adjusts angle based on `crt_rand`).
                jitter = float(int(rand()) % 100 - 50) * 0.06 * max(entry.intensity, 0.0) * dt
                if style == 0:
                    jitter *= 1.96
                    speed = 82.0
                elif style == 8:
                    jitter *= 1.1
                    speed = 62.0
                else:
                    jitter *= 1.1
                    speed = 82.0
                entry.angle -= jitter
                entry.vel = Vec2.from_angle(entry.angle) * speed

            alpha = clamp(entry.intensity, 0.0, 1.0)
            shade = 1.0 - max(entry.intensity, 0.0) * 0.95
            entry.age = alpha
            entry.scale_x = shade
            entry.scale_y = shade
            # Native only updates scale_x/scale_y; scale_z stays at its spawn value (1.0).

            alive = entry.intensity > (0.0 if style == 0 else 0.8)
            if not alive:
                entry.active = False
                expired.append(idx)
                if style == 8 and entry.target_id != -1:
                    target_id = int(entry.target_id)
                    entry.target_id = -1
                    if kill_creature is not None:
                        kill_creature(target_id, int(entry.owner_id))
                    elif creatures is not None and 0 <= target_id < len(creatures):
                        creatures[target_id].hp = -1.0
                        creatures[target_id].active = False
                continue

            if style == 8 and (not entry.render_flag) and entry.target_id != -1 and creatures is not None:
                target_id = int(entry.target_id)
                if 0 <= target_id < len(creatures) and creatures[target_id].active:
                    entry.pos = creatures[target_id].pos

            if entry.render_flag and creatures is not None:
                hit_idx = _creature_find_in_radius(pos=entry.pos, radius=max(entry.intensity, 0.0) * 8.0)
                if hit_idx != -1:
                    entry.render_flag = False
                    creature = creatures[hit_idx]
                    if style == 8:
                        entry.target_id = int(hit_idx)
                        entry.pos = creature.pos
                        entry.vel = Vec2()
                    else:
                        entry.angle = float(entry.angle) % math.tau
                        hit_angle = Vec2(
                            (entry.pos.x - entry.vel.x * dt) - creature.pos.x,
                            (entry.pos.y - entry.vel.y * dt) - creature.pos.y,
                        ).to_angle()
                        hit_angle = float(hit_angle) % math.tau
                        deflect_step = math.tau * 0.2
                        if float(entry.angle) <= float(hit_angle):
                            entry.angle += deflect_step
                        else:
                            entry.angle -= deflect_step

                        bounce_velocity = Vec2.from_angle(float(entry.angle)) * 82.0
                        speed_scale = float(int(rand()) % 10) * 0.1
                        entry.vel = bounce_velocity * speed_scale

                        damage = max(0.0, float(entry.intensity) * 10.0)
                        if damage > 0.0:
                            if apply_creature_damage is not None:
                                apply_creature_damage(int(hit_idx), float(damage), 4, Vec2(), int(entry.owner_id))
                            else:
                                creature.hp -= float(damage)

                        tint = creature.tint
                        tint_sum = float(tint.r) + float(tint.g) + float(tint.b)
                        if tint_sum > 1.6:
                            factor = 1.0 - float(entry.intensity) * 0.01
                            creature.tint = tint.scaled(factor).clamped()

                        if sprite_effects is not None and (idx % 3 == 0):
                            sprite_vel = Vec2(
                                float(int(rand()) % 0x3C - 0x1E),
                                float(int(rand()) % 0x3C - 0x1E),
                            )
                            sprite_effects.spawn(
                                pos=creature.pos,
                                vel=sprite_vel,
                                scale=13.0,
                                color=RGBA(1.0, 1.0, 1.0, 0.7),
                            )

                        if fx_queue is not None:
                            fx_queue.add_random(
                                pos=creature.pos,
                                rand=rand,
                            )

                        creature.pos = creature.pos + entry.vel * dt

        return expired


@dataclass(slots=True)
class SpriteEffect:
    active: bool = False
    color: RGBA = field(default_factory=lambda: RGBA(a=0.0))
    rotation: float = 0.0
    pos: Vec2 = field(default_factory=Vec2)
    vel: Vec2 = field(default_factory=Vec2)
    scale: float = 1.0


class SpriteEffectPool:
    def __init__(self, *, size: int = SPRITE_EFFECT_POOL_SIZE, rand: Callable[[], int] | None = None) -> None:
        self._entries = [SpriteEffect() for _ in range(int(size))]
        self._rand = rand or _default_rand

    @property
    def entries(self) -> list[SpriteEffect]:
        return self._entries

    def reset(self) -> None:
        for entry in self._entries:
            entry.active = False

    def spawn(self, *, pos: Vec2, vel: Vec2, scale: float = 1.0, color: RGBA | None = None) -> int:
        """Port of `fx_spawn_sprite` (0x0041fbb0)."""

        idx = None
        for i, entry in enumerate(self._entries):
            if not entry.active:
                idx = i
                break
        if idx is None:
            if not self._entries:
                raise ValueError("Sprite effect pool has zero entries")
            idx = int(self._rand()) % len(self._entries)

        entry = self._entries[idx]
        entry.active = True
        entry.color = RGBA() if color is None else color
        entry.rotation = float(int(self._rand()) % 0x274) * 0.01
        entry.pos = pos
        entry.vel = vel
        entry.scale = float(scale)
        return idx

    def iter_active(self) -> list[SpriteEffect]:
        return [entry for entry in self._entries if entry.active]

    def update(self, dt: float) -> list[int]:
        if dt <= 0.0:
            return []

        expired: list[int] = []
        for idx, entry in enumerate(self._entries):
            if not entry.active:
                continue
            entry.pos = entry.pos + entry.vel * dt
            entry.rotation += dt * 3.0
            entry.color = entry.color.with_alpha(entry.color.a - dt)
            entry.scale += dt * 60.0
            if entry.color.a <= 0.0:
                entry.active = False
                expired.append(idx)
        return expired


@dataclass(slots=True)
class FxQueueEntry:
    effect_id: int = 0
    rotation: float = 0.0
    pos: Vec2 = field(default_factory=Vec2)
    height: float = 0.0
    width: float = 0.0
    color: RGBA = field(default_factory=RGBA)


class FxQueue:
    """Per-frame terrain decal queue (`fx_queue` / `fx_queue_add`)."""

    def __init__(self, *, capacity: int = FX_QUEUE_CAPACITY, max_count: int = FX_QUEUE_MAX_COUNT) -> None:
        capacity = max(0, int(capacity))
        max_count = max(0, min(int(max_count), capacity))
        self._entries = [FxQueueEntry() for _ in range(capacity)]
        self._count = 0
        self._max_count = max_count

    @property
    def entries(self) -> list[FxQueueEntry]:
        return self._entries

    @property
    def count(self) -> int:
        return self._count

    def clear(self) -> None:
        self._count = 0

    def iter_active(self) -> list[FxQueueEntry]:
        return self._entries[: self._count]

    def add(
        self,
        *,
        effect_id: int,
        pos: Vec2,
        width: float,
        height: float,
        rotation: float,
        rgba: RGBA,
    ) -> bool:
        """Port of `fx_queue_add` (0x0041e840)."""

        if self._count >= self._max_count:
            return False

        entry = self._entries[self._count]
        entry.effect_id = int(effect_id)
        entry.rotation = float(rotation)
        entry.pos = pos
        entry.height = float(height)
        entry.width = float(width)
        entry.color = rgba
        self._count += 1
        return True

    def add_random(self, *, pos: Vec2, rand: Callable[[], int]) -> bool:
        """Port of `fx_queue_add_random` (effect ids 3..7 with grayscale tint)."""
        # Native `fx_queue_add_random` always consumes RNG even when the queue
        # is full, then lets `fx_queue_add` fail silently.
        gray = float(int(rand()) & 0xF) * 0.01 + 0.84
        w = float(int(rand()) % 0x18 - 0x0C) + 30.0
        rotation = float(int(rand()) % 0x274) * 0.01
        effect_id = int(rand()) % 5 + 3
        return self.add(
            effect_id=effect_id,
            pos=pos,
            width=w,
            height=w,
            rotation=rotation,
            rgba=RGBA(gray, gray, gray, 1.0),
        )


@dataclass(slots=True)
class FxQueueRotatedEntry:
    top_left: Vec2 = field(default_factory=Vec2)
    color: RGBA = field(default_factory=RGBA)
    rotation: float = 0.0
    scale: float = 1.0
    creature_type_id: int = 0


class FxQueueRotated:
    """Rotated corpse queue (`fx_queue_rotated` / `fx_queue_add_rotated`)."""

    def __init__(self, *, capacity: int = FX_QUEUE_ROTATED_CAPACITY, max_count: int = FX_QUEUE_ROTATED_MAX_COUNT) -> None:
        capacity = max(0, int(capacity))
        max_count = max(0, min(int(max_count), capacity))
        self._entries = [FxQueueRotatedEntry() for _ in range(capacity)]
        self._count = 0
        self._max_count = max_count

    @property
    def entries(self) -> list[FxQueueRotatedEntry]:
        return self._entries

    @property
    def count(self) -> int:
        return self._count

    def clear(self) -> None:
        self._count = 0

    def iter_active(self) -> list[FxQueueRotatedEntry]:
        return self._entries[: self._count]

    def add(
        self,
        *,
        top_left: Vec2,
        rgba: RGBA,
        rotation: float,
        scale: float,
        creature_type_id: int,
        terrain_bodies_transparency: float = 0.0,
        terrain_texture_failed: bool = False,
    ) -> bool:
        """Port of `fx_queue_add_rotated` (0x00427840)."""

        if terrain_texture_failed:
            return False
        if self._count >= self._max_count:
            return False

        color = rgba
        a = color.a
        if terrain_bodies_transparency != 0.0:
            a = a / float(terrain_bodies_transparency)
        else:
            a = a * 0.8

        entry = self._entries[self._count]
        entry.top_left = top_left
        entry.color = color.with_alpha(a)
        entry.rotation = float(rotation)
        entry.scale = float(scale)
        entry.creature_type_id = int(creature_type_id)

        self._count += 1
        return True


@dataclass(slots=True)
class EffectEntry:
    pos: Vec2 = field(default_factory=Vec2)
    effect_id: int = 0
    vel: Vec2 = field(default_factory=Vec2)
    rotation: float = 0.0
    scale: float = 1.0
    half_width: float = 0.0
    half_height: float = 0.0
    age: float = 0.0
    lifetime: float = 0.0
    flags: int = 0
    color: RGBA = field(default_factory=RGBA)
    rotation_step: float = 0.0
    scale_step: float = 0.0


class EffectPool:
    """Effect pool (`effect_spawn`, `effects_update`).

    This pool renders transient particle quads and can optionally enqueue decals
    into `FxQueue` on expiry (flags bit `0x80`).
    """

    def __init__(self, *, size: int = EFFECT_POOL_SIZE) -> None:
        size = max(0, int(size))
        self._entries = [EffectEntry() for _ in range(size)]
        self._free = list(range(size - 1, -1, -1))
        self._detail_toggle = 0
        self._overwrite_cursor = 0

    @property
    def entries(self) -> list[EffectEntry]:
        return self._entries

    def reset(self) -> None:
        for entry in self._entries:
            entry.flags = 0
        self._free = list(range(len(self._entries) - 1, -1, -1))
        self._detail_toggle = 0
        self._overwrite_cursor = 0

    def iter_active(self) -> list[EffectEntry]:
        return [entry for entry in self._entries if entry.flags]

    def _alloc_slot(self, *, detail_preset: int) -> int | None:
        # Native: if detail_preset < 3, skip every other spawn attempt.
        if int(detail_preset) < 3:
            skip = self._detail_toggle & 1
            self._detail_toggle += 1
            if skip:
                return None

        if self._free:
            return self._free.pop()

        if not self._entries:
            return None

        idx = self._overwrite_cursor % len(self._entries)
        self._overwrite_cursor = idx + 1
        return idx

    def spawn(
        self,
        *,
        effect_id: int,
        pos: Vec2,
        vel: Vec2,
        rotation: float,
        scale: float,
        half_width: float,
        half_height: float,
        age: float,
        lifetime: float,
        flags: int,
        color: RGBA,
        rotation_step: float,
        scale_step: float,
        detail_preset: int,
    ) -> int | None:
        idx = self._alloc_slot(detail_preset=int(detail_preset))
        if idx is None:
            return None

        entry = self._entries[idx]
        entry.pos = pos
        entry.effect_id = int(effect_id)
        entry.vel = vel
        entry.rotation = float(rotation)
        entry.scale = float(scale)
        entry.half_width = float(half_width)
        entry.half_height = float(half_height)
        entry.age = float(age)
        entry.lifetime = float(lifetime)
        entry.flags = int(flags)
        entry.color = color
        entry.rotation_step = float(rotation_step)
        entry.scale_step = float(scale_step)
        return idx

    def free(self, idx: int) -> None:
        if not (0 <= idx < len(self._entries)):
            return
        entry = self._entries[idx]
        entry.flags = 0
        self._free.append(idx)

    def update(self, dt: float, *, fx_queue: FxQueue | None = None) -> None:
        """Advance active effects and enqueue terrain decals on expiry."""

        if dt <= 0.0:
            return

        for idx, entry in enumerate(self._entries):
            flags = int(entry.flags)
            if not flags:
                continue

            age = float(entry.age) + float(dt)
            entry.age = age
            lifetime = float(entry.lifetime)

            if age < lifetime:
                if age >= 0.0:
                    entry.pos = entry.pos + entry.vel * float(dt)
                    if flags & 0x4:
                        entry.rotation += float(entry.rotation_step) * float(dt)
                    if flags & 0x8:
                        entry.scale += float(entry.scale_step) * float(dt)
                    if flags & 0x10:
                        next_alpha = 1.0 - age / lifetime if lifetime > 1e-9 else 0.0
                        entry.color = entry.color.with_alpha(next_alpha)
                continue

            if fx_queue is not None and (flags & 0x80):
                # On expiry, the native code overrides alpha before queuing.
                alpha = 0.35 if (flags & 0x100) else 0.8
                fx_queue.add(
                    effect_id=int(entry.effect_id),
                    pos=entry.pos,
                    width=float(entry.half_width) * 2.0,
                    height=float(entry.half_height) * 2.0,
                    rotation=float(entry.rotation),
                    rgba=entry.color.with_alpha(alpha),
                )

            self.free(idx)

    def spawn_shell_casing(
        self,
        *,
        pos: Vec2,
        aim_heading: float,
        weapon_flags: int,
        rand: Callable[[], int],
        detail_preset: int,
    ) -> None:
        """Port of the casing spawn in `player_update` (effect_id 0x12)."""

        if (int(weapon_flags) & 0x1) == 0:
            return

        angle = float(aim_heading) + float(int(rand()) & 0x3F) * 0.01
        speed = float(int(rand()) & 0x3F) * 0.022727273 + 1.0
        velocity = Vec2.from_angle(angle) * (speed * 100.0)

        rotation = float((int(rand()) & 0x3F) - 0x20) * 0.1
        rotation_step = (float(int(rand()) % 0x14) * 0.1 - 1.0) * 14.0

        self.spawn(
            effect_id=int(EffectId.CASING),
            pos=pos,
            vel=velocity,
            rotation=float(rotation),
            scale=1.0,
            half_width=2.0,
            half_height=2.0,
            age=0.0,
            lifetime=0.15,
            flags=0x1C5,
            color=RGBA(1.0, 1.0, 1.0, 0.6),
            rotation_step=float(rotation_step),
            scale_step=0.0,
            detail_preset=int(detail_preset),
        )

    def spawn_blood_splatter(
        self,
        *,
        pos: Vec2,
        angle: float,
        age: float,
        rand: Callable[[], int],
        detail_preset: int,
        fx_toggle: int,
    ) -> None:
        """Port of `effect_spawn_blood_splatter` (0x0042eb10)."""

        if int(fx_toggle) != 0:
            return

        lifetime = 0.25 - float(age)
        base = float(angle) + math.pi
        direction = Vec2.from_angle(base)

        for _ in range(2):
            r0 = int(rand())
            rotation = float((r0 & 0x3F) - 0x20) * 0.1 + base
            r1 = int(rand())
            half = float((r1 & 7) + 1)
            r2 = int(rand())
            speed_x = float((r2 & 0x3F) + 100)
            r3 = int(rand())
            speed_y = float((r3 & 0x3F) + 100)
            velocity = Vec2(direction.x * speed_x, direction.y * speed_y)
            r4 = int(rand())
            scale_step = float(r4 & 0x7F) * 0.03 + 0.1

            self.spawn(
                effect_id=int(EffectId.BLOOD_SPLATTER),
                pos=pos,
                vel=velocity,
                rotation=rotation,
                scale=1.0,
                half_width=half,
                half_height=half,
                age=float(age),
                lifetime=lifetime,
                flags=0xC9,
                color=RGBA(1.0, 1.0, 1.0, 0.5),
                rotation_step=0.0,
                scale_step=scale_step,
                detail_preset=int(detail_preset),
            )

    def spawn_burst(
        self,
        *,
        pos: Vec2,
        count: int,
        rand: Callable[[], int],
        detail_preset: int,
        lifetime: float = 0.5,
        scale_step: float | None = None,
        color: RGBA = RGBA(0.4, 0.5, 1.0, 0.5),
    ) -> None:
        """Port of `effect_spawn_burst` (0x0042ef60)."""

        count = max(0, int(count))
        for _ in range(count):
            r0 = int(rand())
            rotation = float(r0 & 0x7F) * 0.049087387
            r1 = int(rand())
            vx = float((r1 & 0x7F) - 0x40)
            r2 = int(rand())
            vy = float((r2 & 0x7F) - 0x40)
            velocity = Vec2(vx, vy)
            if scale_step is None:
                r3 = int(rand())
                step = float(r3 % 100) * 0.01 + 0.1
            else:
                step = float(scale_step)

            self.spawn(
                effect_id=int(EffectId.BURST),
                pos=pos,
                vel=velocity,
                rotation=rotation,
                scale=1.0,
                half_width=32.0,
                half_height=32.0,
                age=0.0,
                lifetime=float(lifetime),
                flags=0x1D,
                color=color,
                rotation_step=0.0,
                scale_step=step,
                detail_preset=int(detail_preset),
            )

    def spawn_ring(
        self,
        *,
        pos: Vec2,
        detail_preset: int,
        color: RGBA,
        lifetime: float = 0.25,
        scale_step: float = 50.0,
    ) -> None:
        """Ring/halo burst used by bonus pickup effects (`bonus_apply`)."""

        self.spawn(
            effect_id=int(EffectId.RING),
            pos=pos,
            vel=Vec2(),
            rotation=0.0,
            scale=1.0,
            half_width=32.0,
            half_height=32.0,
            age=0.0,
            lifetime=float(lifetime),
            flags=0x19,
            color=color,
            rotation_step=0.0,
            scale_step=float(scale_step),
            detail_preset=int(detail_preset),
        )

    def spawn_freeze_shard(
        self,
        *,
        pos: Vec2,
        angle: float,
        rand: Callable[[], int],
        detail_preset: int,
    ) -> None:
        """Port of `effect_spawn_freeze_shard` (0x0042ec80)."""

        lifetime = float(int(rand()) & 0xF) * 0.01 + 0.2
        base = float(angle) + math.pi

        rotation = float(int(rand()) % 100) * 0.01 + base
        half = float(int(rand()) % 5 + 7)

        velocity = Vec2.from_angle(base) * 114.0

        rotation_step = (float(int(rand()) % 0x14) * 0.1 - 1.0) * 4.0
        scale_step = -float(int(rand()) & 0xF) * 0.1

        effect_id = int(rand()) % 3 + 8
        self.spawn(
            effect_id=int(effect_id),
            pos=pos,
            vel=velocity,
            rotation=float(rotation),
            scale=1.0,
            half_width=float(half),
            half_height=float(half),
            age=0.0,
            lifetime=float(lifetime),
            flags=0x1CD,
            color=RGBA(1.0, 1.0, 1.0, 0.5),
            rotation_step=float(rotation_step),
            scale_step=float(scale_step),
            detail_preset=int(detail_preset),
        )

    def spawn_freeze_shatter(
        self,
        *,
        pos: Vec2,
        angle: float,
        rand: Callable[[], int],
        detail_preset: int,
    ) -> None:
        """Port of `effect_spawn_freeze_shatter` (0x0042ee00)."""

        lifetime = 1.1
        for idx in range(4):
            rotation = float(idx) * (math.pi / 2.0) + float(angle)
            velocity = Vec2.from_angle(rotation) * 42.0
            half = float(int(rand()) % 10 + 0x12)
            rotation_step = (float(int(rand()) % 0x14) * 0.1 - 1.0) * 1.9

            self.spawn(
                effect_id=int(EffectId.FREEZE_SHATTER),
                pos=pos,
                vel=velocity,
                rotation=float(rotation),
                scale=1.0,
                half_width=float(half),
                half_height=float(half),
                age=0.0,
                lifetime=float(lifetime),
                flags=0x5D,
                color=RGBA(1.0, 1.0, 1.0, 0.5),
                rotation_step=float(rotation_step),
                scale_step=0.0,
                detail_preset=int(detail_preset),
            )

        for _ in range(4):
            shard_angle = float(int(rand()) % 0x264) * 0.01
            self.spawn_freeze_shard(
                pos=pos,
                angle=float(shard_angle),
                rand=rand,
                detail_preset=int(detail_preset),
            )

    def spawn_explosion_burst(
        self,
        *,
        pos: Vec2,
        scale: float,
        rand: Callable[[], int],
        detail_preset: int,
    ) -> None:
        """Port of `effect_spawn_explosion_burst` (0x0042f6c0)."""

        detail_preset = int(detail_preset)
        scale = float(scale)

        # Shockwave ring.
        self.spawn(
            effect_id=int(EffectId.RING),
            pos=pos,
            vel=Vec2(),
            rotation=0.0,
            scale=1.0,
            half_width=32.0,
            half_height=32.0,
            age=-0.1,
            lifetime=0.35,
            flags=0x19,
            color=RGBA(0.6, 0.6, 0.6, 1.0),
            rotation_step=0.0,
            scale_step=scale * 25.0,
            detail_preset=detail_preset,
        )

        # Dark explosion puffs (high detail only).
        if detail_preset > 3:
            for idx in range(2):
                age = float(idx) * 0.2 - 0.5
                lifetime = float(idx) * 0.2 + 0.6
                rotation = float(int(rand()) % 0x266) * 0.02
                self.spawn(
                    effect_id=int(EffectId.EXPLOSION_PUFF),
                    pos=pos,
                    vel=Vec2(),
                    rotation=float(rotation),
                    scale=1.0,
                    half_width=32.0,
                    half_height=32.0,
                    age=float(age),
                    lifetime=float(lifetime),
                    flags=0x5D,
                    color=RGBA(0.1, 0.1, 0.1, 1.0),
                    rotation_step=1.4,
                    scale_step=scale * 5.0,
                    detail_preset=detail_preset,
                )

        # Bright flash.
        self.spawn(
            effect_id=int(EffectId.BURST),
            pos=pos,
            vel=Vec2(),
            rotation=0.0,
            scale=1.0,
            half_width=32.0,
            half_height=32.0,
            age=0.0,
            lifetime=0.3,
            flags=0x19,
            color=RGBA(1.0, 1.0, 1.0, 1.0),
            rotation_step=0.0,
            scale_step=scale * 45.0,
            detail_preset=detail_preset,
        )

        if detail_preset < 2:
            count = 1
        else:
            count = 3 + (1 if detail_preset > 3 else 0)

        # Extra shockwave particles.
        for _ in range(count):
            rotation = float(int(rand()) % 0x13A) * 0.02
            velocity = Vec2(
                float((int(rand()) & 0x3F) * 2 - 0x40),
                float((int(rand()) & 0x3F) * 2 - 0x40),
            )
            scale_step = float((int(rand()) - 3) & 7) * scale
            rotation_step = float((int(rand()) + 3) & 7)
            self.spawn(
                effect_id=int(EffectId.EXPLOSION_BURST),
                pos=pos,
                vel=velocity,
                rotation=float(rotation),
                scale=1.0,
                half_width=32.0,
                half_height=32.0,
                age=0.0,
                lifetime=0.7,
                flags=0x1D,
                color=RGBA(1.0, 1.0, 1.0, 1.0),
                rotation_step=float(rotation_step),
                scale_step=float(scale_step),
                detail_preset=detail_preset,
            )



from __future__ import annotations

from dataclasses import dataclass

DEMO_TOTAL_PLAY_TIME_MS = 2_400_000
DEMO_QUEST_GRACE_TIME_MS = 300_000


def format_demo_trial_time(ms: int) -> str:
    value = int(ms)
    if value < 0:
        value = 0
    minutes = value // 60_000
    seconds = (value // 1_000) % 60
    centiseconds = (value % 1_000) // 10
    return f"{minutes}:{seconds:02d}.{centiseconds:02d}"


@dataclass(frozen=True, slots=True)
class DemoTrialOverlayInfo:
    visible: bool
    kind: str  # "none" | "quest_tier_limit" | "quest_grace_left" | "time_up"
    remaining_ms: int
    remaining_label: str


def tick_demo_trial_timers(
    *,
    demo_build: bool,
    game_mode_id: int,
    overlay_visible: bool,
    global_playtime_ms: int,
    quest_grace_elapsed_ms: int,
    dt_ms: int,
) -> tuple[int, int]:
    """Advance demo timers by `dt_ms` (ms), returning updated values.

    This mirrors the classic behavior where:
      - global playtime accumulates until it hits `DEMO_TOTAL_PLAY_TIME_MS`, then clamps
      - once global time is exhausted, a quest-only grace timer becomes active
      - timers do not advance while the demo trial overlay is shown
    """

    if not demo_build:
        return int(global_playtime_ms), int(quest_grace_elapsed_ms)

    if int(game_mode_id) == 8:
        return int(global_playtime_ms), int(quest_grace_elapsed_ms)

    used_ms = max(0, int(global_playtime_ms))
    grace_ms = max(0, int(quest_grace_elapsed_ms))
    if used_ms >= DEMO_TOTAL_PLAY_TIME_MS and grace_ms < 1:
        grace_ms = 1
    used_ms = min(DEMO_TOTAL_PLAY_TIME_MS, used_ms)

    if overlay_visible:
        return int(used_ms), int(grace_ms)

    delta_ms = int(dt_ms)
    if delta_ms <= 0:
        return int(used_ms), int(grace_ms)

    used_ms = min(DEMO_TOTAL_PLAY_TIME_MS, used_ms + delta_ms)
    if used_ms >= DEMO_TOTAL_PLAY_TIME_MS and grace_ms < 1:
        grace_ms = 1

    if grace_ms > 0 and int(game_mode_id) == 3:
        grace_ms += delta_ms

    return int(used_ms), int(grace_ms)


def demo_trial_overlay_info(
    *,
    demo_build: bool,
    game_mode_id: int,
    global_playtime_ms: int,
    quest_grace_elapsed_ms: int,
    quest_stage_major: int,
    quest_stage_minor: int,
) -> DemoTrialOverlayInfo:
    """Compute demo trial overlay status.

    Modeled after `demo_trial_overlay_render` (0x004047c0) call sites and time formatting.

    Notes:
      - `global_playtime_ms` maps to `game_status_blob.game_sequence_id` (ms).
      - `quest_grace_elapsed_ms` maps to `demo_trial_elapsed_ms` (ms) once activated.
    """

    if not demo_build:
        return DemoTrialOverlayInfo(False, "none", 0, format_demo_trial_time(0))

    mode_id = int(game_mode_id)
    if mode_id == 8:  # tutorial
        return DemoTrialOverlayInfo(False, "none", 0, format_demo_trial_time(0))

    used_ms = max(0, int(global_playtime_ms))
    grace_ms = max(0, int(quest_grace_elapsed_ms))

    global_remaining_ms = max(0, DEMO_TOTAL_PLAY_TIME_MS - used_ms)
    grace_remaining_ms = max(0, DEMO_QUEST_GRACE_TIME_MS - grace_ms)

    tier_locked = mode_id == 3 and (int(quest_stage_major) > 1 or int(quest_stage_minor) > 10)

    if grace_ms > 0:
        if grace_remaining_ms <= 0:
            return DemoTrialOverlayInfo(True, "time_up", 0, format_demo_trial_time(0))
        if tier_locked:
            return DemoTrialOverlayInfo(
                True,
                "quest_tier_limit",
                int(grace_remaining_ms),
                format_demo_trial_time(grace_remaining_ms),
            )
        # During the quest-only grace period, the classic demo blocks other modes
        # and points the player back to Quests.
        if mode_id != 3:
            return DemoTrialOverlayInfo(
                True,
                "quest_grace_left",
                int(grace_remaining_ms),
                format_demo_trial_time(grace_remaining_ms),
            )
        return DemoTrialOverlayInfo(False, "none", int(grace_remaining_ms), format_demo_trial_time(grace_remaining_ms))

    if global_remaining_ms <= 0:
        return DemoTrialOverlayInfo(True, "time_up", 0, format_demo_trial_time(0))

    # Demo tier gating: the classic demo lets you play stage 1 quests only; once
    # the player reaches stage > 1, it shows the upsell overlay even if time remains.
    if tier_locked:
        return DemoTrialOverlayInfo(
            True,
            "quest_tier_limit",
            int(global_remaining_ms),
            format_demo_trial_time(global_remaining_ms),
        )

    return DemoTrialOverlayInfo(False, "none", int(global_remaining_ms), format_demo_trial_time(global_remaining_ms))



from __future__ import annotations

import io
import inspect
import json
import random
import re
from pathlib import Path
from dataclasses import fields, is_dataclass
from typing import Any, cast

import typer
from PIL import Image

from grim import jaz, paq
from grim.geom import Vec2
from grim.rand import Crand
from .paths import default_runtime_dir
from .creatures.spawn import SpawnEnv, build_spawn_plan, spawn_id_label
from .quests import all_quests
from .quests.types import QuestContext, QuestDefinition, SpawnEntry


app = typer.Typer(add_completion=False)
replay_app = typer.Typer(add_completion=False)
original_app = typer.Typer(add_completion=False)
app.add_typer(replay_app, name="replay")
app.add_typer(original_app, name="original")

_QUEST_DEFS: dict[str, QuestDefinition] = {quest.level: quest for quest in all_quests()}
_QUEST_BUILDERS = {level: quest.builder for level, quest in _QUEST_DEFS.items()}
_QUEST_TITLES = {level: quest.title for level, quest in _QUEST_DEFS.items()}

_SEP_RE = re.compile(r"[\\/]+")


def _safe_relpath(name: str) -> Path:
    parts = [p for p in _SEP_RE.split(name) if p]
    if not parts:
        raise ValueError("empty entry name")
    for part in parts:
        if part in (".", ".."):
            raise ValueError(f"unsafe path part: {part!r}")
    return Path(*parts)


def _extract_one(paq_path: Path, assets_root: Path) -> int:
    out_root = assets_root / paq_path.stem
    out_root.mkdir(parents=True, exist_ok=True)
    count = 0
    for name, data in paq.iter_entries(paq_path):
        rel = _safe_relpath(name)
        dest = out_root / rel
        dest.parent.mkdir(parents=True, exist_ok=True)
        suffix = dest.suffix.lower()
        if suffix == ".jaz":
            jaz_image = jaz.decode_jaz_bytes(data)
            base = dest.with_suffix("")
            jaz_image.composite_image().save(base.with_suffix(".png"))
        else:
            if suffix == ".tga":
                img = Image.open(io.BytesIO(data))
                img.save(dest.with_suffix(".png"))
            else:
                dest.write_bytes(data)
        count += 1
    return count


@app.command("extract")
def cmd_extract(game_dir: Path, assets_dir: Path) -> None:
    """Extract all .paq files into a flat asset directory."""
    if not game_dir.is_dir():
        typer.echo(f"game dir not found: {game_dir}", err=True)
        raise typer.Exit(code=1)
    assets_dir.mkdir(parents=True, exist_ok=True)
    paqs = sorted(game_dir.rglob("*.paq"))
    if not paqs:
        typer.echo(f"no .paq files under {game_dir}", err=True)
        raise typer.Exit(code=1)
    total = 0
    for paq_path in paqs:
        total += _extract_one(paq_path, assets_dir)
    typer.echo(f"extracted {total} files")


def _call_builder(builder, ctx: QuestContext, rng: random.Random | None) -> list[SpawnEntry]:
    params = inspect.signature(builder).parameters
    if "rng" in params:
        return builder(ctx, rng=rng)
    return builder(ctx)


def _format_entry(idx: int, entry: SpawnEntry, *, plan_info: tuple[int, int] | None) -> str:
    creature = spawn_id_label(entry.spawn_id)
    plan_text = ""
    if plan_info is not None:
        creatures_per_spawn, spawn_slots_per_spawn = plan_info
        alloc = entry.count * creatures_per_spawn
        plan_text = f"  alloc={alloc:3d} (x{creatures_per_spawn:2d})  slots={spawn_slots_per_spawn}"
    return (
        f"{idx:02d}  t={entry.trigger_ms:5d}  "
        f"id=0x{entry.spawn_id:02x} ({entry.spawn_id:2d})  "
        f"creature={creature:10s}  "
        f"count={entry.count:2d}  "
        f"x={entry.pos.x:7.1f}  y={entry.pos.y:7.1f}  heading={entry.heading:7.3f}{plan_text}"
    )


def _format_id(value: int | None) -> str:
    if value is None:
        return "none"
    return f"0x{value:02x} ({value})"


def _format_id_list(values: tuple[int, ...] | None) -> str:
    if not values:
        return "none"
    return "[" + ", ".join(_format_id(value) for value in values) + "]"


def _format_meta(quest: QuestDefinition) -> list[str]:
    builder_addr = f"0x{quest.builder_address:08x}" if quest.builder_address is not None else "unknown"
    terrain_ids = _format_id_list(quest.terrain_ids)
    return [
        f"time_limit_ms={quest.time_limit_ms}",
        f"start_weapon_id={quest.start_weapon_id}",
        f"unlock_perk_id={_format_id(quest.unlock_perk_id)}",
        f"unlock_weapon_id={_format_id(quest.unlock_weapon_id)}",
        f"builder_address={builder_addr}",
        f"terrain_ids={terrain_ids}",
    ]


@app.command("quests")
def cmd_quests(
    level: str = typer.Argument(..., help="quest level, e.g. 1.1"),
    width: int = typer.Option(1024, help="terrain width"),
    height: int = typer.Option(1024, help="terrain height"),
    player_count: int = typer.Option(1, help="player count"),
    seed: int | None = typer.Option(None, help="seed for randomized quests"),
    sort: bool = typer.Option(False, help="sort output by trigger time"),
    show_plan: bool = typer.Option(False, help="include spawn-plan allocation summary"),
) -> None:
    """Print quest spawn scripts for a given level."""
    quest = _QUEST_DEFS.get(level)
    if quest is None:
        available = ", ".join(sorted(_QUEST_BUILDERS))
        typer.echo(f"unknown level {level!r}. Available: {available}", err=True)
        raise typer.Exit(code=1)
    builder = quest.builder
    title = quest.title
    ctx = QuestContext(width=width, height=height, player_count=player_count)
    rng = random.Random(seed) if seed is not None else random.Random()
    entries = _call_builder(builder, ctx, rng)
    if sort:
        entries = sorted(entries, key=lambda e: (e.trigger_ms, e.spawn_id, e.x, e.y))
    typer.echo(f"Quest {level} {title} ({len(entries)} entries)")
    typer.echo("Meta: " + "; ".join(_format_meta(quest)))

    plan_cache: dict[int, tuple[int, int]] = {}
    if show_plan:
        env = SpawnEnv(
            terrain_width=float(width),
            terrain_height=float(height),
            demo_mode_active=True,
            hardcore=False,
            difficulty_level=0,
        )
        for entry in entries:
            if entry.spawn_id in plan_cache:
                continue
            plan = build_spawn_plan(entry.spawn_id, Vec2(512.0, 512.0), 0.0, Crand(0), env)
            plan_cache[entry.spawn_id] = (len(plan.creatures), len(plan.spawn_slots))
        total_alloc = sum(entry.count * plan_cache[entry.spawn_id][0] for entry in entries)
        total_slots = sum(entry.count * plan_cache[entry.spawn_id][1] for entry in entries)
        typer.echo(f"Plan: total_alloc={total_alloc} total_spawn_slots={total_slots}")

    for idx, entry in enumerate(entries, start=1):
        typer.echo(_format_entry(idx, entry, plan_info=plan_cache.get(entry.spawn_id)))


@app.command("view")
def cmd_view(
    name: str = typer.Argument(..., help="view name (e.g. empty)"),
    width: int = typer.Option(1024, help="window width"),
    height: int = typer.Option(768, help="window height"),
    fps: int = typer.Option(60, help="target fps"),
    preserve_bugs: bool = typer.Option(False, "--preserve-bugs", help="preserve known original exe bugs/quirks"),
    assets_dir: Path = typer.Option(Path("artifacts") / "assets", help="assets root (default: ./artifacts/assets)"),
) -> None:
    """Launch a Raylib debug view."""
    from grim.app import run_view
    from grim.view import ViewContext
    from .views import all_views, view_by_name

    view_def = view_by_name(name)
    if view_def is None:
        available = ", ".join(view.name for view in all_views())
        typer.echo(f"unknown view {name!r}. Available: {available}", err=True)
        raise typer.Exit(code=1)
    ctx = ViewContext(assets_dir=assets_dir, preserve_bugs=bool(preserve_bugs))
    params = inspect.signature(view_def.factory).parameters
    if "ctx" in params:
        view = view_def.factory(ctx=ctx)
    else:
        view = view_def.factory()
    title = f"{view_def.title} — Crimsonland"
    run_view(view, width=width, height=height, title=title, fps=fps)


@replay_app.command("play")
def cmd_replay_play(
    replay_file: Path = typer.Argument(..., help="replay file path (.crdemo.gz)"),
    width: int | None = typer.Option(None, help="window width (default: use crimson.cfg)"),
    height: int | None = typer.Option(None, help="window height (default: use crimson.cfg)"),
    fps: int = typer.Option(60, help="target fps"),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
    assets_dir: Path | None = typer.Option(
        None,
        help="assets root (default: base-dir; missing .paq files are downloaded)",
    ),
) -> None:
    """Play back a recorded replay."""
    from grim.app import run_view
    from grim.console import create_console
    from grim.config import ensure_crimson_cfg
    from grim.view import ViewContext

    from .assets_fetch import download_missing_paqs
    from .modes.replay_playback_mode import ReplayPlaybackMode

    if assets_dir is None:
        assets_dir = base_dir
    base_dir.mkdir(parents=True, exist_ok=True)
    cfg = ensure_crimson_cfg(base_dir)
    if width is None:
        width = int(cfg.screen_width)
    if height is None:
        height = int(cfg.screen_height)
    console = create_console(base_dir, assets_dir=assets_dir)
    download_missing_paqs(assets_dir, console)

    ctx = ViewContext(assets_dir=assets_dir, preserve_bugs=False)
    view = ReplayPlaybackMode(ctx, replay_path=replay_file)
    title = f"Replay — {replay_file.name}"
    run_view(view, width=width, height=height, title=title, fps=fps)


@replay_app.command("verify")
def cmd_replay_verify(
    replay_file: Path = typer.Argument(..., help="replay file path (.crdemo.gz)"),
    checkpoints_file: Path | None = typer.Option(
        None,
        "--checkpoints",
        help="checkpoint sidecar path (default: <replay>.checkpoints.json.gz)",
    ),
    max_ticks: int | None = typer.Option(None, help="stop after N ticks (default: full replay)"),
    strict_events: bool = typer.Option(
        False,
        "--strict-events/--lenient-events",
        help="fail on unsupported replay events/perk picks (default: lenient)",
    ),
    trace_rng: bool = typer.Option(
        False,
        "--trace-rng",
        help="include presentation RNG draw marks in verification checkpoints",
    ),
) -> None:
    """Verify a replay by comparing headless checkpoints with a sidecar file."""
    import hashlib

    from .game_modes import GameMode
    from .original.diff import compare_checkpoints
    from .replay import load_replay
    from .replay.checkpoints import default_checkpoints_path, load_checkpoints_file
    from .sim.runners import ReplayRunnerError, run_rush_replay, run_survival_replay

    replay_bytes = Path(replay_file).read_bytes()
    replay_sha256 = hashlib.sha256(replay_bytes).hexdigest()
    replay = load_replay(replay_bytes)

    if checkpoints_file is None:
        checkpoints_file = default_checkpoints_path(replay_file)
    checkpoints_path = Path(checkpoints_file)
    if not checkpoints_path.is_file():
        typer.echo(f"checkpoints file not found: {checkpoints_path}", err=True)
        raise typer.Exit(code=1)

    expected = load_checkpoints_file(checkpoints_path)
    if expected.replay_sha256 and str(expected.replay_sha256) != str(replay_sha256):
        typer.echo(
            f"warning: checkpoints replay_sha256 mismatch (checkpoints={expected.replay_sha256!r}, replay={replay_sha256!r})",
            err=True,
        )

    checkpoint_ticks = {int(ckpt.tick_index) for ckpt in expected.checkpoints}
    actual = []

    mode = int(replay.header.game_mode_id)
    try:
        if mode == int(GameMode.SURVIVAL):
            result = run_survival_replay(
                replay,
                max_ticks=max_ticks,
                strict_events=bool(strict_events),
                trace_rng=bool(trace_rng),
                checkpoints_out=actual,
                checkpoint_ticks=checkpoint_ticks,
            )
        elif mode == int(GameMode.RUSH):
            result = run_rush_replay(
                replay,
                max_ticks=max_ticks,
                trace_rng=bool(trace_rng),
                checkpoints_out=actual,
                checkpoint_ticks=checkpoint_ticks,
            )
        else:
            typer.echo(f"unsupported replay game_mode_id={mode}", err=True)
            raise typer.Exit(code=1)
    except ReplayRunnerError as exc:
        typer.echo(f"replay verification failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    diff = compare_checkpoints(expected.checkpoints, actual)
    if not diff.ok:
        failure = diff.failure
        assert failure is not None
        exp = failure.expected
        act = failure.actual

        if failure.kind == "missing_checkpoint":
            typer.echo(f"checkpoint missing at tick={int(failure.tick_index)}", err=True)
            raise typer.Exit(code=1)

        if failure.kind == "command_mismatch":
            assert act is not None
            typer.echo(f"checkpoint command mismatch at tick={int(failure.tick_index)}", err=True)
            typer.echo(f"  command_hash expected={exp.command_hash} actual={act.command_hash}", err=True)
            if int(exp.events.hit_count) >= 0:
                typer.echo(
                    "  events "
                    f"expected=(hits={exp.events.hit_count}, pickups={exp.events.pickup_count}, sfx={exp.events.sfx_count}, head={exp.events.sfx_head}) "
                    f"actual=(hits={act.events.hit_count}, pickups={act.events.pickup_count}, sfx={act.events.sfx_count}, head={act.events.sfx_head})",
                    err=True,
                )
            raise typer.Exit(code=1)

        assert act is not None
        typer.echo(f"checkpoint mismatch at tick={int(failure.tick_index)}", err=True)
        typer.echo(f"  state_hash expected={exp.state_hash} actual={act.state_hash}", err=True)
        typer.echo(f"  rng_state expected={exp.rng_state} actual={act.rng_state}", err=True)
        typer.echo(f"  score_xp expected={exp.score_xp} actual={act.score_xp}", err=True)
        typer.echo(f"  kills expected={exp.kills} actual={act.kills}", err=True)
        typer.echo(f"  creature_count expected={exp.creature_count} actual={act.creature_count}", err=True)
        typer.echo(f"  perk_pending expected={exp.perk_pending} actual={act.perk_pending}", err=True)
        if failure.first_rng_mark is not None:
            key = str(failure.first_rng_mark)
            typer.echo(
                f"  rng_mark[{key}] expected={exp.rng_marks.get(key)} actual={act.rng_marks.get(key)}",
                err=True,
            )
        typer.echo(f"  deaths expected={len(exp.deaths)} actual={len(act.deaths)}", err=True)
        if exp.deaths or act.deaths:
            typer.echo(f"  first death expected={exp.deaths[:1]} actual={act.deaths[:1]}", err=True)
        if int(exp.events.hit_count) >= 0:
            typer.echo(
                "  events "
                f"expected=(hits={exp.events.hit_count}, pickups={exp.events.pickup_count}, sfx={exp.events.sfx_count}, head={exp.events.sfx_head}) "
                f"actual=(hits={act.events.hit_count}, pickups={act.events.pickup_count}, sfx={act.events.sfx_count}, head={act.events.sfx_head})",
                err=True,
            )
        if exp.perk != act.perk:
            typer.echo(
                "  perk snapshot differs "
                f"(expected pending={exp.perk.pending_count} choices={exp.perk.choices}, "
                f"actual pending={act.perk.pending_count} choices={act.perk.choices})",
                err=True,
            )
        raise typer.Exit(code=1)

    message = (
        f"ok: {len(expected.checkpoints)} checkpoints match; ticks={result.ticks} "
        f"score_xp={result.score_xp} kills={result.creature_kill_count}"
    )
    if diff.first_rng_only_tick is not None:
        message += f"; rng-only drift starts at tick={diff.first_rng_only_tick}"
    typer.echo(message)


@replay_app.command("diff-checkpoints")
def cmd_replay_diff_checkpoints(
    expected_file: Path = typer.Argument(..., help="expected checkpoints sidecar (.json.gz)"),
    actual_file: Path = typer.Argument(..., help="actual checkpoints sidecar (.json.gz)"),
) -> None:
    """Compare two checkpoint sidecars and report the first divergence."""
    from .replay.checkpoints import load_checkpoints_file
    from .original.diff import compare_checkpoints

    expected = load_checkpoints_file(Path(expected_file))
    actual = load_checkpoints_file(Path(actual_file))
    diff = compare_checkpoints(expected.checkpoints, actual.checkpoints)
    if not diff.ok:
        failure = diff.failure
        assert failure is not None
        exp = failure.expected
        act = failure.actual

        if failure.kind == "missing_checkpoint":
            typer.echo(f"checkpoint missing at tick={int(failure.tick_index)}", err=True)
            raise typer.Exit(code=1)

        if failure.kind == "command_mismatch":
            assert act is not None
            typer.echo(f"checkpoint command mismatch at tick={int(failure.tick_index)}", err=True)
            typer.echo(f"  command_hash expected={exp.command_hash} actual={act.command_hash}", err=True)
            if int(exp.events.hit_count) >= 0:
                typer.echo(
                    "  events "
                    f"expected=(hits={exp.events.hit_count}, pickups={exp.events.pickup_count}, sfx={exp.events.sfx_count}, head={exp.events.sfx_head}) "
                    f"actual=(hits={act.events.hit_count}, pickups={act.events.pickup_count}, sfx={act.events.sfx_count}, head={act.events.sfx_head})",
                    err=True,
                )
            raise typer.Exit(code=1)

        assert act is not None
        typer.echo(f"checkpoint mismatch at tick={int(failure.tick_index)}", err=True)
        typer.echo(f"  state_hash expected={exp.state_hash} actual={act.state_hash}", err=True)
        typer.echo(f"  rng_state expected={exp.rng_state} actual={act.rng_state}", err=True)
        typer.echo(f"  score_xp expected={exp.score_xp} actual={act.score_xp}", err=True)
        typer.echo(f"  kills expected={exp.kills} actual={act.kills}", err=True)
        typer.echo(f"  creature_count expected={exp.creature_count} actual={act.creature_count}", err=True)
        typer.echo(f"  perk_pending expected={exp.perk_pending} actual={act.perk_pending}", err=True)
        if failure.first_rng_mark is not None:
            key = str(failure.first_rng_mark)
            typer.echo(
                f"  rng_mark[{key}] expected={exp.rng_marks.get(key)} actual={act.rng_marks.get(key)}",
                err=True,
            )
        typer.echo(f"  deaths expected={len(exp.deaths)} actual={len(act.deaths)}", err=True)
        if exp.deaths or act.deaths:
            typer.echo(f"  first death expected={exp.deaths[:1]} actual={act.deaths[:1]}", err=True)
        if int(exp.events.hit_count) >= 0:
            typer.echo(
                "  events "
                f"expected=(hits={exp.events.hit_count}, pickups={exp.events.pickup_count}, sfx={exp.events.sfx_count}, head={exp.events.sfx_head}) "
                f"actual=(hits={act.events.hit_count}, pickups={act.events.pickup_count}, sfx={act.events.sfx_count}, head={act.events.sfx_head})",
                err=True,
            )
        if exp.perk != act.perk:
            typer.echo(
                "  perk snapshot differs "
                f"(expected pending={exp.perk.pending_count} choices={exp.perk.choices}, "
                f"actual pending={act.perk.pending_count} choices={act.perk.choices})",
                err=True,
            )
        raise typer.Exit(code=1)

    message = f"ok: {len(expected.checkpoints)} checkpoints match"
    if diff.first_rng_only_tick is not None:
        message += f"; rng-only drift starts at tick={diff.first_rng_only_tick}"
    typer.echo(message)


@original_app.command("verify-capture")
def cmd_replay_verify_capture(
    capture_file: Path = typer.Argument(
        ...,
        help="capture file (.json/.json.gz)",
    ),
    max_ticks: int | None = typer.Option(None, help="stop after N ticks (default: full capture)"),
    seed: int | None = typer.Option(
        None,
        help="seed override for replay reconstruction (default: infer from capture rng telemetry)",
    ),
    strict_events: bool = typer.Option(
        False,
        "--strict-events/--lenient-events",
        help="fail on unsupported replay events/perk picks (default: lenient)",
    ),
    trace_rng: bool = typer.Option(
        False,
        "--trace-rng",
        help="include presentation RNG draw marks in generated checkpoints",
    ),
    max_field_diffs: int = typer.Option(
        16,
        "--max-field-diffs",
        min=1,
        help="max field-level diffs to print for first mismatching tick",
    ),
    float_abs_tol: float = typer.Option(
        0.001,
        "--float-abs-tol",
        min=0.0,
        help="absolute tolerance for float field comparisons",
    ),
) -> None:
    """Verify capture ticks directly against rewrite simulation state."""
    from .original.capture import load_capture
    from .original.verify import (
        CaptureVerifyError,
        verify_capture,
    )
    from .sim.runners import ReplayRunnerError

    capture = load_capture(Path(capture_file))

    try:
        result, run_result = verify_capture(
            capture,
            seed=seed,
            max_ticks=max_ticks,
            strict_events=bool(strict_events),
            trace_rng=bool(trace_rng),
            max_field_diffs=int(max_field_diffs),
            float_abs_tol=float(float_abs_tol),
        )
    except (ReplayRunnerError, CaptureVerifyError) as exc:
        typer.echo(f"capture verification failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    if not result.ok:
        failure = result.failure
        assert failure is not None
        typer.echo(f"capture mismatch at tick={int(failure.tick_index)}", err=True)
        typer.echo(
            "  note: compared checkpoint state fields only "
            "(ignoring command_hash/state_hash/rng_state/rng_marks domains)",
            err=True,
        )
        if result.elapsed_baseline_tick is not None and result.elapsed_offset_ms is not None:
            typer.echo(
                "  elapsed baseline "
                f"tick={result.elapsed_baseline_tick} offset_ms(actual-expected)={result.elapsed_offset_ms}",
                err=True,
            )

        if failure.kind == "missing_checkpoint":
            typer.echo("  checkpoint missing in rewrite output", err=True)
            typer.echo(
                f"  checked={result.checked_count}/{result.expected_count} actual_samples={result.actual_count}",
                err=True,
            )
            typer.echo(
                "  run_result "
                f"ticks={run_result.ticks} score_xp={run_result.score_xp} kills={run_result.creature_kill_count}",
                err=True,
            )
            raise typer.Exit(code=1)

        assert failure.actual is not None
        if not failure.field_diffs:
            typer.echo("  mismatch detected but no detailed field diff was collected", err=True)
        else:
            for field_diff in failure.field_diffs:
                typer.echo(
                    f"  {field_diff.field}: expected={field_diff.expected!r} actual={field_diff.actual!r}",
                    err=True,
                )
        typer.echo(
            "  run_result "
            f"ticks={run_result.ticks} score_xp={run_result.score_xp} kills={run_result.creature_kill_count}",
            err=True,
        )
        raise typer.Exit(code=1)

    message = (
        f"ok: {result.checked_count} capture checkpoints match "
        f"(state fields); ticks={run_result.ticks} score_xp={run_result.score_xp} "
        f"kills={run_result.creature_kill_count}"
    )
    if result.elapsed_baseline_tick is not None and result.elapsed_offset_ms is not None:
        message += (
            f"; elapsed baseline tick={result.elapsed_baseline_tick} "
            f"offset_ms(actual-expected)={result.elapsed_offset_ms}"
        )
    typer.echo(message)


@original_app.command("convert-capture")
def cmd_replay_convert_capture(
    capture_file: Path = typer.Argument(
        ...,
        help="capture file (.json/.json.gz)",
    ),
    output_file: Path = typer.Argument(..., help="output checkpoints sidecar (.json.gz)"),
    replay_file: Path | None = typer.Option(
        None,
        "--replay",
        help="output replay path (.crdemo.gz); default: derive from checkpoints path",
    ),
    replay_sha256: str = typer.Option(
        "",
        help="optional replay sha256 to store in the converted sidecar",
    ),
    seed: int | None = typer.Option(
        None,
        help="seed override for replay reconstruction (default: infer from capture rng telemetry)",
    ),
) -> None:
    """Convert capture data into replay + checkpoint artifacts."""
    import hashlib

    from .replay import dump_replay
    from .replay.checkpoints import dump_checkpoints_file
    from .original.capture import (
        convert_capture_to_checkpoints,
        convert_capture_to_replay,
        default_capture_replay_path,
        load_capture,
    )

    capture = load_capture(Path(capture_file))
    replay = convert_capture_to_replay(capture, seed=seed)
    replay_path = (
        Path(replay_file) if replay_file is not None else default_capture_replay_path(Path(output_file))
    )
    replay_blob = dump_replay(replay)
    replay_path.write_bytes(replay_blob)
    digest = hashlib.sha256(replay_blob).hexdigest()

    checkpoints = convert_capture_to_checkpoints(
        capture,
        replay_sha256=str(replay_sha256 or digest),
    )
    dump_checkpoints_file(Path(output_file), checkpoints)
    typer.echo(f"wrote replay ({len(replay.inputs)} ticks) to {replay_path}")
    typer.echo(f"wrote {len(checkpoints.checkpoints)} checkpoints to {output_file}")
    typer.echo("note: replay uses best-effort input reconstruction; checkpoints remain the authoritative diff target")


@original_app.command(
    "divergence-report",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True, "help_option_names": []},
)
def cmd_replay_divergence_report(ctx: typer.Context) -> None:
    """Run divergence report against a capture."""
    from .original import divergence_report

    raise typer.Exit(code=divergence_report.main(list(ctx.args)))


@original_app.command(
    "focus-trace",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True, "help_option_names": []},
)
def cmd_replay_focus_trace(ctx: typer.Context) -> None:
    """Trace a single focus tick from capture diagnostics."""
    from .original import focus_trace

    raise typer.Exit(code=focus_trace.main(list(ctx.args)))


@original_app.command(
    "creature-trajectory",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True, "help_option_names": []},
)
def cmd_replay_creature_trajectory(ctx: typer.Context) -> None:
    """Trace capture-vs-rewrite creature trajectory drift."""
    from .original import creature_trajectory

    raise typer.Exit(code=creature_trajectory.main(list(ctx.args)))


@app.callback(invoke_without_command=True)
def cmd_game(
    ctx: typer.Context,
    width: int | None = typer.Option(None, help="window width (default: use crimson.cfg)"),
    height: int | None = typer.Option(None, help="window height (default: use crimson.cfg)"),
    fps: int = typer.Option(60, help="target fps"),
    seed: int | None = typer.Option(None, help="rng seed"),
    demo: bool = typer.Option(False, "--demo", help="enable shareware demo mode"),
    no_intro: bool = typer.Option(False, "--no-intro", help="skip company splashes and intro music"),
    debug: bool = typer.Option(False, "--debug", help="enable debug cheats and overlays"),
    preserve_bugs: bool = typer.Option(False, "--preserve-bugs", help="preserve known original exe bugs/quirks"),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
    assets_dir: Path | None = typer.Option(
        None,
        help="assets root (default: base-dir; missing .paq files are downloaded)",
    ),
) -> None:
    """Run the reimplementation game flow (default command)."""
    if ctx.invoked_subcommand:
        return
    from .game import GameConfig, run_game

    config = GameConfig(
        base_dir=base_dir,
        assets_dir=assets_dir,
        width=width,
        height=height,
        fps=fps,
        seed=seed,
        demo_enabled=demo,
        no_intro=no_intro,
        debug=debug,
        preserve_bugs=bool(preserve_bugs),
    )
    run_game(config)


@app.command("config")
def cmd_config(
    path: Path | None = typer.Option(None, help="path to crimson.cfg (default: base-dir/crimson.cfg)"),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
) -> None:
    """Inspect crimson.cfg configuration values."""
    from grim.config import CRIMSON_CFG_NAME, CRIMSON_CFG_STRUCT, load_crimson_cfg

    cfg_path = path if path is not None else base_dir / CRIMSON_CFG_NAME
    config = load_crimson_cfg(cfg_path)
    typer.echo(f"path: {config.path}")
    typer.echo(f"screen: {config.screen_width}x{config.screen_height}")
    typer.echo(f"windowed: {config.windowed_flag}")
    typer.echo(f"bpp: {config.screen_bpp}")
    typer.echo(f"texture_scale: {config.texture_scale}")
    typer.echo("fields:")
    for sub in CRIMSON_CFG_STRUCT.subcons:
        name = sub.name
        if not name:
            continue
        value = config.data[name]
        typer.echo(f"{name}: {_format_cfg_value(value)}")


def _format_cfg_value(value: object) -> str:
    if isinstance(value, (bytes, bytearray)):
        length = len(value)
        prefix = value.split(b"\x00", 1)[0]
        if prefix and all(32 <= b < 127 for b in prefix):
            text = prefix.decode("ascii", errors="replace")
            return f"{text!r} (len={length})"
        return f"0x{bytes(value).hex()} (len={length})"
    return str(value)


def _parse_int_auto(text: str) -> int:
    try:
        return int(text, 0)
    except ValueError as exc:
        raise typer.BadParameter(f"invalid integer: {text!r}") from exc


def _parse_vec2(text: str) -> Vec2:
    raw = text.strip()
    if "," in raw:
        left, right = raw.split(",", 1)
    else:
        parts = raw.split()
        if len(parts) != 2:
            raise typer.BadParameter(f"invalid vec2: {text!r} (expected 'x,y' or 'x y')")
        left, right = parts
    try:
        return Vec2(float(left.strip()), float(right.strip()))
    except ValueError as exc:
        raise typer.BadParameter(f"invalid vec2: {text!r}") from exc


def _dc_to_dict(obj: object) -> dict[str, object]:
    if not is_dataclass(obj):
        raise TypeError(f"expected dataclass instance, got {type(obj).__name__}")
    dc_obj = cast(Any, obj)
    return {f.name: getattr(dc_obj, f.name) for f in fields(dc_obj)}


@app.command("spawn-plan")
def cmd_spawn_plan(
    template: str = typer.Argument(..., help="spawn id (e.g. 0x12)"),
    seed: str = typer.Option("0xBEEF", help="MSVCRT rand() seed (e.g. 0xBEEF)"),
    pos: str = typer.Option("512,512", help="spawn position as 'x,y'"),
    heading: float = typer.Option(0.0, help="heading (radians)"),
    terrain_w: float = typer.Option(1024.0, help="terrain width"),
    terrain_h: float = typer.Option(1024.0, help="terrain height"),
    demo_mode_active: bool = typer.Option(True, help="when true, burst effect is skipped"),
    hardcore: bool = typer.Option(False, help="hardcore mode"),
    difficulty: int = typer.Option(0, help="difficulty level"),
    as_json: bool = typer.Option(False, "--json", help="print JSON"),
) -> None:
    """Build and print a spawn plan for a single template id."""
    template_id = _parse_int_auto(template)
    rng = Crand(_parse_int_auto(seed))
    spawn_pos = _parse_vec2(pos)
    env = SpawnEnv(
        terrain_width=terrain_w,
        terrain_height=terrain_h,
        demo_mode_active=demo_mode_active,
        hardcore=hardcore,
        difficulty_level=difficulty,
    )
    plan = build_spawn_plan(template_id, spawn_pos, heading, rng, env)
    if as_json:
        payload: dict[str, object] = {
            "template_id": template_id,
            "pos": [spawn_pos.x, spawn_pos.y],
            "heading": heading,
            "seed": _parse_int_auto(seed),
            "env": {
                "terrain_width": terrain_w,
                "terrain_height": terrain_h,
                "demo_mode_active": demo_mode_active,
                "hardcore": hardcore,
                "difficulty_level": difficulty,
            },
            "primary": plan.primary,
            "creatures": [_dc_to_dict(c) for c in plan.creatures],
            "spawn_slots": [_dc_to_dict(s) for s in plan.spawn_slots],
            "effects": [_dc_to_dict(e) for e in plan.effects],
            "rng_state": rng.state,
        }
        typer.echo(json.dumps(payload, indent=2, sort_keys=True))
        return

    typer.echo(f"template_id=0x{template_id:02x} ({template_id}) creature={spawn_id_label(template_id)}")
    typer.echo(
        f"pos=({spawn_pos.x:.1f},{spawn_pos.y:.1f}) "
        f"heading={heading:.6f} seed=0x{_parse_int_auto(seed):08x} rng_state=0x{rng.state:08x}"
    )
    typer.echo(
        "env="
        f"demo_mode_active={demo_mode_active} "
        f"hardcore={hardcore} "
        f"difficulty={difficulty} "
        f"terrain={terrain_w:.0f}x{terrain_h:.0f}"
    )
    typer.echo(f"primary={plan.primary} creatures={len(plan.creatures)} slots={len(plan.spawn_slots)} effects={len(plan.effects)}")
    typer.echo("")
    typer.echo("creatures:")
    for idx, c in enumerate(plan.creatures):
        primary = "*" if idx == plan.primary else " "
        typer.echo(
            f"{primary}{idx:02d} type={c.type_id!s:14s} ai={c.ai_mode:2d} flags=0x{int(c.flags):03x} "
            f"pos=({c.pos.x:7.1f},{c.pos.y:7.1f}) health={c.health!s:>6s} size={c.size!s:>6s} link={c.ai_link_parent!s:>3s} "
            f"slot={c.spawn_slot!s:>3s}"
        )
    if plan.spawn_slots:
        typer.echo("")
        typer.echo("spawn_slots:")
        for idx, slot in enumerate(plan.spawn_slots):
            typer.echo(
                f"{idx:02d} owner={slot.owner_creature:02d} timer={slot.timer:.2f} count={slot.count:3d} "
                f"limit={slot.limit:3d} interval={slot.interval:.3f} child=0x{slot.child_template_id:02x}"
            )
    if plan.effects:
        typer.echo("")
        typer.echo("effects:")
        for fx in plan.effects:
            typer.echo(f"burst x={fx.pos.x:.1f} y={fx.pos.y:.1f} count={fx.count}")


@app.command("oracle")
def cmd_oracle(
    seed: int = typer.Option(0xBEEF, help="RNG seed for deterministic runs"),
    input_file: Path | None = typer.Option(None, "--input-file", "-i", help="JSON file with input sequence"),
    max_frames: int = typer.Option(36000, help="Maximum frames to run (default: 10 min at 60fps)"),
    frame_rate: int = typer.Option(60, help="Frame rate for simulation"),
    sample_rate: int = typer.Option(60, "--sample-rate", "-s", help="Emit state every N frames (1=every frame, 60=1/sec)"),
    preserve_bugs: bool = typer.Option(False, "--preserve-bugs", help="preserve known original exe bugs/quirks"),
    output_mode: str = typer.Option(
        "summary",
        "--output", "-o",
        help="Output mode: full (all entities), summary (fast), hash (ultra-fast), checkpoints (on events only)",
    ),
) -> None:
    """Run headless oracle mode for differential testing.

    Emits JSON game state to stdout. Use with --seed for deterministic runs
    and --input-file for replaying specific input sequences.

    Output modes:
      - summary: Score, kills, player pos/health (default, fast)
      - full: All entities including creatures, projectiles, bonuses
      - hash: SHA256 hash of full state (ultra-fast comparison)
      - checkpoints: Emit only when score/kills/level/weapon changes

    Examples:
        # Fast validation at 1 Hz sampling
        crimson oracle --seed 12345 -i replay.json -s 60 -o summary

        # Full frame-by-frame for debugging divergence
        crimson oracle --seed 12345 -i replay.json -s 1 -o full

        # Ultra-fast hash comparison
        crimson oracle --seed 12345 -i replay.json -o hash

        # Event-driven checkpoints only
        crimson oracle --seed 12345 -i replay.json -o checkpoints
    """
    from .oracle import OracleConfig, OutputMode, run_headless

    # Validate output mode
    mode_map = {
        "full": OutputMode.FULL,
        "summary": OutputMode.SUMMARY,
        "hash": OutputMode.HASH,
        "checkpoints": OutputMode.CHECKPOINTS,
    }
    if output_mode not in mode_map:
        typer.echo(f"Invalid output mode: {output_mode!r}. Choose from: {', '.join(mode_map)}", err=True)
        raise typer.Exit(code=1)

    config = OracleConfig(
        seed=seed,
        input_file=input_file,
        max_frames=max_frames,
        frame_rate=frame_rate,
        sample_rate=sample_rate,
        output_mode=mode_map[output_mode],
        preserve_bugs=bool(preserve_bugs),
    )
    run_headless(config)


def main(argv: list[str] | None = None) -> None:
    app(prog_name="crimson", args=argv)


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path
import random
from typing import Any, Protocol


class PauseBackground(Protocol):
    def draw_pause_background(self) -> None: ...


class GameState(Protocol):
    # Keep this protocol lightweight: frontend code should not depend on the full
    # gameplay/sim stack (enforced via import-linter).
    base_dir: Path
    assets_dir: Path
    rng: random.Random
    config: Any
    status: Any
    preserve_bugs: bool

    texture_cache: Any
    audio: Any
    resource_paq: Path
    logos: Any
    console: Any

    menu_ground: Any
    menu_ground_camera: Any
    pause_background: PauseBackground | None

    demo_enabled: bool
    skip_intro: bool
    menu_sign_locked: bool

    quit_requested: bool
    screen_fade_alpha: float
    screen_fade_ramp: bool



from __future__ import annotations

import struct
from typing import Sequence

from ..gameplay import PlayerInput
from .types import (
    PackedPlayerInput,
    PackedTickInputs,
    PerkMenuOpenEvent,
    PerkPickEvent,
    Replay,
    ReplayEvent,
    ReplayHeader,
    pack_input_flags,
)

_FORMAT_VERSION = 1


def _quantize_f32(value: float) -> float:
    return struct.unpack("<f", struct.pack("<f", float(value)))[0]


class ReplayRecorder:
    def __init__(self, header: ReplayHeader, *, version: int = _FORMAT_VERSION) -> None:
        if int(version) != _FORMAT_VERSION:
            raise ValueError(f"unsupported replay version: {version}")
        self._version = int(version)
        self._header = header
        self._tick_index = 0
        self._inputs: list[PackedTickInputs] = []
        self._events: list[ReplayEvent] = []

    @property
    def header(self) -> ReplayHeader:
        return self._header

    @property
    def tick_index(self) -> int:
        return int(self._tick_index)

    def record_tick(self, inputs: Sequence[PlayerInput]) -> int:
        """Record a single simulation tick worth of inputs.

        Returns the tick index that was recorded.
        """

        player_count = int(self._header.player_count)
        if len(inputs) != player_count:
            raise ValueError(f"expected {player_count} player inputs, got {len(inputs)}")

        quant = self._header.input_quantization
        tick: PackedTickInputs = []
        for inp in inputs:
            mx = inp.move.x
            my = inp.move.y
            ax = inp.aim.x
            ay = inp.aim.y
            if quant == "f32":
                mx = _quantize_f32(mx)
                my = _quantize_f32(my)
                ax = _quantize_f32(ax)
                ay = _quantize_f32(ay)
            flags = pack_input_flags(
                fire_down=bool(inp.fire_down),
                fire_pressed=bool(inp.fire_pressed),
                reload_pressed=bool(inp.reload_pressed),
            )
            packed_input: PackedPlayerInput = [mx, my, [ax, ay], flags]
            tick.append(packed_input)

        tick_index = int(self._tick_index)
        self._inputs.append(tick)
        self._tick_index += 1
        return tick_index

    def record_perk_pick(
        self,
        *,
        player_index: int,
        choice_index: int,
        tick_index: int | None = None,
    ) -> None:
        if tick_index is None:
            tick_index = int(self._tick_index)
        self._events.append(
            PerkPickEvent(
                tick_index=int(tick_index),
                player_index=int(player_index),
                choice_index=int(choice_index),
            )
        )

    def record_perk_menu_open(
        self,
        *,
        player_index: int,
        tick_index: int | None = None,
    ) -> None:
        if tick_index is None:
            tick_index = int(self._tick_index)
        self._events.append(
            PerkMenuOpenEvent(
                tick_index=int(tick_index),
                player_index=int(player_index),
            )
        )

    def finish(self) -> Replay:
        return Replay(
            version=int(self._version),
            header=self._header,
            inputs=self._inputs,
            events=list(self._events),
        )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import pyray as rl

from crimson.creatures.anim import creature_corpse_frame_for_type
from crimson.creatures.spawn import CreatureTypeId
from grim.assets import resolve_asset_path
from grim.config import ensure_crimson_cfg
from grim.fonts.small import SmallFontData, load_small_font
from grim.geom import Vec2
from grim.terrain_render import GroundCorpseDecal, GroundRenderer, _maybe_alpha_test
from grim.view import View, ViewContext

from ..paths import default_runtime_dir
from ._ui_helpers import draw_ui_text, ui_line_height
from .registry import register_view


WORLD_SIZE = 1024.0
WINDOW_W = 1024
WINDOW_H = 768

BG = rl.Color(235, 235, 235, 255)
UI_TEXT = rl.Color(20, 20, 20, 255)
UI_HINT = rl.Color(70, 70, 70, 255)
UI_ERROR = rl.Color(240, 80, 80, 255)


@dataclass(frozen=True, slots=True)
class _Step:
    name: str
    description: str


_STEPS: tuple[_Step, ...] = (
    _Step(name="clear", description="Clear the ground render target"),
    _Step(name="shadow", description="Bake shadow pass only (correct order)"),
    _Step(name="color", description="Bake color pass only (correct order)"),
    _Step(name="clear", description="Clear the ground render target"),
    _Step(name="color", description="Bake color pass only (wrong order)"),
    _Step(name="shadow", description="Bake shadow pass only (wrong order)"),
)


class CorpseStampDebugView:
    def __init__(self, ctx: ViewContext) -> None:
        self._assets_root = ctx.assets_dir
        self._missing_assets: list[str] = []
        self._small: SmallFontData | None = None

        self._owned_textures: list[rl.Texture] = []
        self._ground: GroundRenderer | None = None
        self._bodyset: rl.Texture | None = None

        self.close_requested = False
        self._step_index = 0
        self._corpse_size = 256.0
        self._corpse_rotation = 0.0
        self._screenshot_requested = False
        self._dump_requested = False
        self._dump_index = 0

    def _load_runtime_config(self) -> tuple[float, float | None, float | None]:
        runtime_dir = default_runtime_dir()
        if runtime_dir.is_dir():
            try:
                cfg = ensure_crimson_cfg(runtime_dir)
                return float(cfg.texture_scale), float(cfg.screen_width), float(cfg.screen_height)
            except Exception:
                return 1.0, None, None
        return 1.0, None, None

    def _dump_render_target(self) -> Path | None:
        ground = self._ground
        if ground is None or ground.render_target is None:
            return None

        log_dir = Path("artifacts") / "debug"
        try:
            log_dir.mkdir(parents=True, exist_ok=True)
        except Exception:
            log_dir = Path("artifacts")

        step = _STEPS[self._step_index]
        alpha_test = "a1" if bool(getattr(ground, "alpha_test", True)) else "a0"
        filename = f"corpse_stamp_rt_{self._dump_index:03d}_step{self._step_index + 1:02d}_{step.name}_{alpha_test}.png"
        out_path = log_dir / filename

        image = rl.load_image_from_texture(ground.render_target.texture)
        # Render textures are vertically flipped in raylib.
        rl.image_flip_vertical(image)
        try:
            rl.export_image(image, str(out_path))
        finally:
            rl.unload_image(image)
        self._dump_index += 1
        return out_path

    def _corpse_decal(self) -> GroundCorpseDecal:
        size = float(self._corpse_size)
        frame = creature_corpse_frame_for_type(int(CreatureTypeId.SPIDER_SP1))
        cx = WORLD_SIZE * 0.5
        cy = WORLD_SIZE * 0.5
        return GroundCorpseDecal(
            bodyset_frame=int(frame),
            top_left=Vec2(cx - size * 0.5, cy - size * 0.5),
            size=size,
            rotation_rad=float(self._corpse_rotation),
            tint=rl.Color(255, 255, 255, 255),
        )

    def _clear_ground(self) -> None:
        ground = self._ground
        if ground is None:
            return
        ground.create_render_target()
        if ground.render_target is None:
            return
        rl.begin_texture_mode(ground.render_target)
        rl.clear_background(BG)
        rl.end_texture_mode()
        # GroundRenderer treats this as an internal invariant; set it for debug fills.
        ground._render_target_ready = True

    def _bake_shadow_only(self) -> None:
        ground = self._ground
        bodyset = self._bodyset
        if ground is None or bodyset is None:
            return
        ground.create_render_target()
        if ground.render_target is None:
            return

        scale = ground._normalized_texture_scale()
        inv_scale = 1.0 / scale
        offset = 2.0 * scale / float(ground.width)
        ground._set_texture_filters((bodyset,), point=True)

        rl.begin_texture_mode(ground.render_target)
        with _maybe_alpha_test(ground.alpha_test):
            ground._draw_corpse_shadow_pass(bodyset, [self._corpse_decal()], inv_scale, offset)
        rl.end_texture_mode()

        ground._set_texture_filters((bodyset,), point=False)
        ground._render_target_ready = True

    def _bake_color_only(self) -> None:
        ground = self._ground
        bodyset = self._bodyset
        if ground is None or bodyset is None:
            return
        ground.bake_corpse_decals(bodyset, [self._corpse_decal()], shadow=False)

    def _apply_step(self) -> None:
        step = _STEPS[self._step_index]
        if step.name == "clear":
            self._clear_ground()
        elif step.name == "shadow":
            self._bake_shadow_only()
        elif step.name == "color":
            self._bake_color_only()

    def open(self) -> None:
        rl.set_window_size(WINDOW_W, WINDOW_H)
        self._missing_assets.clear()
        self._owned_textures.clear()
        self._ground = None
        self._bodyset = None
        self.close_requested = False
        self._step_index = 0

        try:
            self._small = load_small_font(self._assets_root, self._missing_assets)
        except Exception:
            self._small = None

        base_path = resolve_asset_path(self._assets_root, "ter/ter_q1_base.png")
        bodyset_path = resolve_asset_path(self._assets_root, "game/bodyset.png")
        if base_path is None:
            self._missing_assets.append("ter/ter_q1_base.png")
        if bodyset_path is None:
            self._missing_assets.append("game/bodyset.png")
        if self._missing_assets:
            raise FileNotFoundError("Missing assets: " + ", ".join(self._missing_assets))

        base = rl.load_texture(str(base_path))
        bodyset = rl.load_texture(str(bodyset_path))
        self._owned_textures.extend([base, bodyset])
        self._bodyset = bodyset

        texture_scale, screen_w, screen_h = self._load_runtime_config()
        self._ground = GroundRenderer(
            texture=base,
            width=int(WORLD_SIZE),
            height=int(WORLD_SIZE),
            texture_scale=float(texture_scale),
            screen_width=screen_w,
            screen_height=screen_h,
        )
        self._ground.alpha_test = True
        self._clear_ground()

    def close(self) -> None:
        if self._ground is not None and self._ground.render_target is not None:
            rl.unload_render_texture(self._ground.render_target)
            self._ground.render_target = None
        self._ground = None
        self._bodyset = None

        for texture in self._owned_textures:
            rl.unload_texture(texture)
        self._owned_textures.clear()

        if self._small is not None:
            rl.unload_texture(self._small.texture)
            self._small = None

    def update(self, dt: float) -> None:
        del dt
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self.close_requested = True
            return

        if rl.is_key_pressed(rl.KeyboardKey.KEY_R):
            self._step_index = 0
            self._clear_ground()

        if rl.is_key_pressed(rl.KeyboardKey.KEY_A):
            if self._ground is not None:
                self._ground.alpha_test = not bool(self._ground.alpha_test)

        if rl.is_key_pressed(rl.KeyboardKey.KEY_N) or rl.is_key_pressed(rl.KeyboardKey.KEY_SPACE):
            self._step_index = (self._step_index + 1) % len(_STEPS)
            self._apply_step()

        if rl.is_key_pressed(rl.KeyboardKey.KEY_P):
            self._screenshot_requested = True

        if rl.is_key_pressed(rl.KeyboardKey.KEY_D):
            self._dump_requested = True

        if rl.is_key_down(rl.KeyboardKey.KEY_Q):
            self._corpse_rotation -= 0.04
        if rl.is_key_down(rl.KeyboardKey.KEY_E):
            self._corpse_rotation += 0.04

    def consume_screenshot_request(self) -> bool:
        requested = self._screenshot_requested
        self._screenshot_requested = False
        return requested

    def draw(self) -> None:
        rl.clear_background(BG)

        if self._missing_assets:
            draw_ui_text(
                self._small, "Missing assets: " + ", ".join(self._missing_assets), Vec2(24, 24), color=UI_ERROR
            )
            return

        ground = self._ground
        if ground is None:
            draw_ui_text(self._small, "Ground renderer not initialized.", Vec2(24, 24), color=UI_ERROR)
            return

        if self._dump_requested:
            self._dump_requested = False
            self._dump_render_target()

        screen_w = float(rl.get_screen_width())
        screen_h = float(rl.get_screen_height())
        camera = Vec2(screen_w * 0.5 - WORLD_SIZE * 0.5, screen_h * 0.5 - WORLD_SIZE * 0.5)
        ground.draw(camera, screen_w=screen_w, screen_h=screen_h)

        # UI
        x = 24.0
        y = 20.0
        line = float(ui_line_height(self._small))
        step = _STEPS[self._step_index]
        alpha_test = bool(getattr(ground, "alpha_test", True))
        draw_ui_text(self._small, "Corpse stamp debug (SPIDER)", Vec2(x, y), color=UI_TEXT)
        y += line
        draw_ui_text(
            self._small,
            "N/Space: next step   R: reset   A: toggle alpha test   Q/E: rotate   P: screenshot   D: dump RT",
            Vec2(x, y),
            color=UI_HINT,
        )
        y += line
        draw_ui_text(
            self._small, f"step {self._step_index + 1}/{len(_STEPS)}: {step.description}", Vec2(x, y), color=UI_HINT
        )
        y += line
        draw_ui_text(
            self._small,
            f"alpha_test={'on' if alpha_test else 'off'}  size={self._corpse_size:.1f}  dump_index={self._dump_index}",
            Vec2(x, y),
            color=UI_HINT,
        )

        # Source preview (bodyset frame) in the corner for inspection.
        if self._bodyset is not None:
            frame = creature_corpse_frame_for_type(int(CreatureTypeId.SPIDER_SP1))
            src = ground._corpse_src(self._bodyset, int(frame))
            preview = 256.0
            pad = 18.0
            dst = rl.Rectangle(screen_w - preview - pad, pad, preview, preview)
            rl.draw_rectangle(int(dst.x) - 2, int(dst.y) - 2, int(preview) + 4, int(preview) + 4, rl.Color(0, 0, 0, 30))
            rl.draw_texture_pro(self._bodyset, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)


@register_view("corpse-stamp-debug", "Corpse stamp debug")
def build_corpse_stamp_debug_view(ctx: ViewContext) -> View:
    return CorpseStampDebugView(ctx)



from __future__ import annotations

from .capture import (
    CAPTURE_BOOTSTRAP_EVENT_KIND,
    CAPTURE_PERK_APPLY_EVENT_KIND,
    CAPTURE_PERK_PENDING_EVENT_KIND,
    CaptureError,
    apply_capture_bootstrap_payload,
    build_capture_dt_frame_ms_i32_overrides,
    build_capture_dt_frame_overrides,
    capture_bootstrap_payload_from_event_payload,
    capture_perk_apply_id_from_event_payload,
    capture_perk_pending_from_event_payload,
    convert_capture_to_checkpoints,
    convert_capture_to_replay,
    default_capture_replay_path,
    dump_capture,
    infer_capture_seed,
    load_capture,
)
from .diff import (
    DEFAULT_RNG_MARK_ORDER,
    ReplayDiffFailure,
    ReplayDiffResult,
    ReplayFieldDiff,
    checkpoint_field_diffs,
    compare_checkpoints,
)
from .schema import (
    CaptureConfig,
    CaptureEventHead,
    CaptureFile,
    CapturePhaseMarker,
    CaptureTick,
    ModuleInfo,
    ProcessInfo,
    SessionFingerprint,
)
__all__ = [
    "CAPTURE_BOOTSTRAP_EVENT_KIND",
    "CAPTURE_PERK_APPLY_EVENT_KIND",
    "CAPTURE_PERK_PENDING_EVENT_KIND",
    "CaptureConfig",
    "CaptureError",
    "CaptureEventHead",
    "CaptureFile",
    "CapturePhaseMarker",
    "CaptureTick",
    "DEFAULT_RNG_MARK_ORDER",
    "ModuleInfo",
    "ProcessInfo",
    "ReplayDiffFailure",
    "ReplayDiffResult",
    "ReplayFieldDiff",
    "SessionFingerprint",
    "apply_capture_bootstrap_payload",
    "build_capture_dt_frame_ms_i32_overrides",
    "build_capture_dt_frame_overrides",
    "capture_bootstrap_payload_from_event_payload",
    "capture_perk_apply_id_from_event_payload",
    "capture_perk_pending_from_event_payload",
    "checkpoint_field_diffs",
    "compare_checkpoints",
    "convert_capture_to_checkpoints",
    "convert_capture_to_replay",
    "default_capture_replay_path",
    "dump_capture",
    "infer_capture_seed",
    "load_capture",
]



from __future__ import annotations

from grim.geom import SupportsXY, Vec2


def mouse_inside_rect_with_padding(
    mouse: SupportsXY,
    *,
    pos: Vec2,
    width: float,
    height: float,
    left_pad: float = 10.0,
    top_pad: float = 2.0,
) -> bool:
    """Port of `ui_mouse_inside_rect_with_padding` (0x00403430)."""

    x = float(pos.x)
    y = float(pos.y)
    return (
        x - float(left_pad) <= float(mouse.x) <= x + float(width)
        and y - float(top_pad) <= float(mouse.y) <= y + float(height)
    )
